# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from functools import partial

import jax
import ml_dtypes

from jax_scalify.core import Array, DTypeLike

from .utils import map_on_backward, map_on_forward


def reduce_precision_dtype_base(arr: Array, dtype: DTypeLike) -> Array:
    """`Fake` cast to an ML dtype (e.g. FP8), using JAX LAX `reduce_precision` operator."""
    info = ml_dtypes.finfo(dtype)
    return jax.lax.reduce_precision(arr, exponent_bits=info.nexp, mantissa_bits=info.nmant)


def reduce_precision_on_forward(arr: Array, dtype: DTypeLike) -> Array:
    """`Fake` cast to an ML dtype, on the forward pass (no-op on backward pass)."""
    return partial(map_on_forward, lambda v: reduce_precision_dtype_base(v, dtype))(arr)


def reduce_precision_on_backward(arr: Array, dtype: DTypeLike) -> Array:
    """`Fake` cast to an ML dtype on the backward pass (no-op on forward pass)."""
    return partial(map_on_backward, lambda v: reduce_precision_dtype_base(v, dtype))(arr)


def cast_on_forward(arr: Array, dtype: DTypeLike) -> Array:
    """Cast input array only on the forward pass (no-op on the backward pass).

    Useful for implementation `DenseGeneral` FP8 matmuls.
    """
    return partial(map_on_forward, lambda v: jax.lax.convert_element_type(v, dtype))(arr)


def cast_on_backward(arr: Array, dtype: DTypeLike) -> Array:
    """Cast input array only on the backward pass (no-op on the forward pass).

    Useful for implementation `DenseGeneral` FP8 matmuls.
    """
    return partial(map_on_backward, lambda v: jax.lax.convert_element_type(v, dtype))(arr)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from typing import Any, Dict, Optional, Sequence, Tuple

import jax.numpy as jnp
import numpy as np
from jax import lax
from jax._src.ad_util import add_any_p

from jax_scalify import core
from jax_scalify.core import DTypeLike, ScaledArray, get_scalify_config, pow2_round, register_scaled_op, safe_div

from .scaled_ops_common import check_scalar_scales, promote_scale_types


def scaled_add_sub(A: ScaledArray, B: ScaledArray, binary_op: Any) -> ScaledArray:
    """Scaled add/sub generic implementation."""
    # TODO: understand when promotion is really required?
    # A, B = as_scaled_array((A, B))  # type:ignore
    check_scalar_scales(A, B)
    A, B = promote_scale_types(A, B)
    assert np.issubdtype(A.scale.dtype, np.floating)
    # Pow2 rounding for unit scaling "rule".
    pow2_rounding_mode = get_scalify_config().rounding_mode
    # TODO: what happens to `sqrt` for non-floating scale?
    # More stable than direct L2 norm, to avoid scale overflow.
    ABscale_max = lax.max(A.scale, B.scale)
    ABscale_min = lax.min(A.scale, B.scale)
    ABscale_ratio = safe_div(ABscale_min, ABscale_max)
    output_scale = ABscale_max * lax.sqrt(1 + ABscale_ratio * ABscale_ratio)
    # Transform back to power-of-2
    output_scale = pow2_round(output_scale, pow2_rounding_mode)
    # Output dtype => promotion of A and B dtypes.
    outdtype = jnp.promote_types(A.dtype, B.dtype)
    Arescale = safe_div(A.scale, output_scale).astype(outdtype)
    Brescale = safe_div(B.scale, output_scale).astype(outdtype)
    # check correct type output if mismatch between data and scale precision
    output_data = binary_op(Arescale * A.data, Brescale * B.data)
    return ScaledArray(output_data, output_scale)


@core.register_scaled_lax_op
def scaled_add(A: ScaledArray, B: ScaledArray) -> ScaledArray:
    return scaled_add_sub(A, B, lax.add)


# TODO: understand difference between `add` and `add_anys`
register_scaled_op(add_any_p, scaled_add)


@core.register_scaled_lax_op
def scaled_sub(A: ScaledArray, B: ScaledArray) -> ScaledArray:
    return scaled_add_sub(A, B, lax.sub)


@core.register_scaled_lax_op
def scaled_dot_general(
    lhs: ScaledArray,
    rhs: ScaledArray,
    dimension_numbers: Tuple[Tuple[Sequence[int], Sequence[int]], Tuple[Sequence[int], Sequence[int]]],
    precision: Any = None,
    preferred_element_type: Optional[DTypeLike] = None,
) -> ScaledArray:
    # Checks on `dot_general` arguments. Only supporting a subset right now.
    ((lhs_contracting_dims, rhs_contracting_dims), (lhs_batch_dims, rhs_batch_dims)) = dimension_numbers
    assert len(lhs_batch_dims) == 0
    assert len(rhs_batch_dims) == 0
    assert len(lhs_contracting_dims) == 1
    assert len(rhs_contracting_dims) == 1

    # Pow2 rounding for unit scaling "rule".
    pow2_rounding_mode = get_scalify_config().rounding_mode
    contracting_dim_size = lhs.shape[lhs_contracting_dims[0]]
    # "unit scaling" rule, based on the contracting axis.
    outscale_dtype = jnp.promote_types(lhs.scale.dtype, rhs.scale.dtype)
    contracting_rescale = np.sqrt(contracting_dim_size).astype(outscale_dtype)
    contracting_rescale = pow2_round(contracting_rescale, pow2_rounding_mode)
    # Keeping power of 2 scale.
    output_scale = lhs.scale * rhs.scale * contracting_rescale.astype(outscale_dtype)
    # NOTE: need to be a bit careful about scale promotion?
    output_data = lax.dot_general(
        lhs.data,
        rhs.data,
        dimension_numbers=dimension_numbers,
        precision=precision,
        preferred_element_type=preferred_element_type,
    )
    output_data = output_data / contracting_rescale.astype(output_data.dtype)
    return ScaledArray(output_data, output_scale)


@core.register_scaled_lax_op
def scaled_conv_general_dilated(lhs: ScaledArray, rhs: ScaledArray, **params: Dict[str, Any]) -> ScaledArray:
    assert isinstance(lhs, ScaledArray)
    assert isinstance(rhs, ScaledArray)
    data = lax.conv_general_dilated_p.bind(lhs.data, rhs.data, **params)
    # FIXME: should we change scaling if e.g. window > 3?
    return ScaledArray(data, lhs.scale * rhs.scale)


@core.register_scaled_lax_op
def scaled_reduce_sum(val: ScaledArray, axes: Tuple[int]) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    shape = val.shape
    scale_dtype = val.scale.dtype
    axes_size = np.array([shape[idx] for idx in axes])
    # Pow2 rounding for unit scaling "rule".
    pow2_rounding_mode = get_scalify_config().rounding_mode
    # Rescale data component following reduction axes & round to power of 2 value.
    axes_rescale = np.sqrt(np.prod(axes_size)).astype(scale_dtype)
    axes_rescale = pow2_round(axes_rescale, pow2_rounding_mode)
    data = lax.reduce_sum_p.bind(val.data, axes=axes) / axes_rescale.astype(val.data.dtype)
    outscale = val.scale * axes_rescale.astype(scale_dtype)
    return ScaledArray(data, outscale)


@core.register_scaled_lax_op
def scaled_reduce_prod(val: ScaledArray, axes: Tuple[int]) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    shape = val.shape
    data = lax.reduce_prod_p.bind(val.data, axes=axes)
    axes_size = np.prod(np.array([shape[idx] for idx in axes]))
    # Stable for power of 2.
    scale = lax.integer_pow(val.scale, axes_size)
    return ScaledArray(data, scale)


@core.register_scaled_lax_op
def scaled_reduce_max(val: ScaledArray, axes: Tuple[int]) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    data = lax.reduce_max_p.bind(val.data, axes=axes)
    # unchanged scaling.
    return ScaledArray(data, val.scale)


@core.register_scaled_lax_op
def scaled_reduce_min(val: ScaledArray, axes: Tuple[int]) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    data = lax.reduce_min_p.bind(val.data, axes=axes)
    # unchanged scaling.
    return ScaledArray(data, val.scale)


@core.register_scaled_lax_op
def scaled_reduce_window_sum(
    val: ScaledArray,
    window_dimensions: Any,
    window_strides: Any,
    padding: Any,
    base_dilation: Any,
    window_dilation: Any,
) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    data = lax.reduce_window_sum_p.bind(
        val.data,
        window_dimensions=window_dimensions,
        window_strides=window_strides,
        padding=padding,
        base_dilation=base_dilation,
        window_dilation=window_dilation,
    )
    # FIXME: should we change scaling if e.g. window > 3?
    return ScaledArray(data, val.scale)


@core.register_scaled_lax_op
def scaled_reduce_window_min(
    val: ScaledArray,
    window_dimensions: Any,
    window_strides: Any,
    padding: Any,
    base_dilation: Any,
    window_dilation: Any,
) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    data = lax.reduce_window_min_p.bind(
        val.data,
        window_dimensions=window_dimensions,
        window_strides=window_strides,
        padding=padding,
        base_dilation=base_dilation,
        window_dilation=window_dilation,
    )
    # unchanged scaling.
    return ScaledArray(data, val.scale)


@core.register_scaled_lax_op
def scaled_reduce_window_max(
    val: ScaledArray,
    window_dimensions: Any,
    window_strides: Any,
    padding: Any,
    base_dilation: Any,
    window_dilation: Any,
) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    data = lax.reduce_window_max_p.bind(
        val.data,
        window_dimensions=window_dimensions,
        window_strides=window_strides,
        padding=padding,
        base_dilation=base_dilation,
        window_dilation=window_dilation,
    )
    # unchanged scaling.
    return ScaledArray(data, val.scale)



# Copyright (c) 2024 Graphcore Ltd. All rights reserved.
import chex
import numpy as np

import jax_scalify as jsa


class ScalifyTreeUtilTests(chex.TestCase):
    def test__tree_flatten__proper_result(self):
        values = {"a": np.int32(2), "b": jsa.as_scaled_array(np.float32(1.5), 1.0)}
        outputs, _ = jsa.tree.flatten(values)
        assert len(outputs) == 2
        assert outputs[0] == 2
        assert isinstance(outputs[1], jsa.ScaledArray)
        assert np.asarray(outputs[1]) == 1.5

    def test__tree_leaves__proper_result(self):
        values = {"a": np.int32(2), "b": jsa.as_scaled_array(np.float32(1.5), 1.0)}
        outputs = jsa.tree.leaves(values)
        assert len(outputs) == 2
        assert outputs[0] == 2
        assert isinstance(outputs[1], jsa.ScaledArray)
        assert np.asarray(outputs[1]) == 1.5

    def test__tree_structure__proper_result(self):
        values = {"a": np.int32(2), "b": jsa.as_scaled_array(np.float32(1.5), 1.0)}
        pytree = jsa.tree.structure(values)
        assert pytree == jsa.tree.flatten(values)[1]

    def test__tree_unflatten__proper_result(self):
        values_in = {"a": np.int32(2), "b": jsa.as_scaled_array(np.float32(1.5), 1.0)}
        outputs, pytree = jsa.tree.flatten(values_in)
        values_out = jsa.tree.unflatten(pytree, outputs)
        assert values_out == values_in

    def test__tree_map__proper_result(self):
        values = {"a": np.int32(2), "b": jsa.as_scaled_array(np.float32(1.5), 1.0)}
        outputs = jsa.tree.map(lambda v: v.dtype, values)
        assert outputs == {"a": np.int32, "b": np.float32}

    def test__tree_astype__all_leaves_casting(self):
        values = {"a": np.int32(2), "b": jsa.as_scaled_array(np.float32(1.5), 1.0)}
        outputs = jsa.tree.astype(values, dtype=np.float16)
        dtypes = jsa.tree.map(lambda v: v.dtype, outputs)
        assert dtypes == {"a": np.float16, "b": np.float16}

    def test__tree_astype__only_float_casting(self):
        values = {"a": np.int32(2), "b": jsa.as_scaled_array(np.float32(1.5), 1.0)}
        outputs = jsa.tree.astype(values, dtype=np.float16, floating_only=True)
        dtypes = jsa.tree.map(lambda v: v.dtype, outputs)
        assert dtypes == {"a": np.int32, "b": np.float16}



# Copyright (c) 2024 Graphcore Ltd. All rights reserved.
from typing import Any, Callable

import jax
import jax.numpy as jnp
from jax import tree_util

from jax_scalify.core import DTypeLike, is_scaled_leaf

Leaf = Any


def astype(tree: Any, dtype: DTypeLike, floating_only: bool = False) -> Any:
    """Map `astype` method to all pytree leaves, `Array` or `ScaledArray`.

    Args:
        tree: the pytree to cast.
        dtype: Dtype to cast to.
        floating_only: Only convert leaves with floating datatype.

    Returns:
        A new PyTree with the same structure, with casting to new dtype.
    """
    if floating_only:
        # Convert only leaves with floating dtype.
        cast_fn = lambda v: v.astype(dtype) if jnp.issubdtype(v.dtype, jnp.floating) else v
        return tree_util.tree_map(cast_fn, tree, is_leaf=is_scaled_leaf)
    return tree_util.tree_map(lambda v: v.astype(dtype), tree, is_leaf=is_scaled_leaf)


def all(tree: Any) -> bool:
    """Call all() over the leaves of a tree, `Array` or `ScaledArray`

    Args:
        tree: the pytree to evaluate
    Returns:
        result: boolean True or False
    """
    return all(jax.tree_util.tree_leaves(tree, is_leaf=is_scaled_leaf))


def flatten(tree: Any) -> tuple[list[Leaf], tree_util.PyTreeDef]:
    """Flattens a pytree, with `Array` or `ScaledArray` leaves.

    The flattening order (i.e. the order of elements in the output list)
    is deterministic, corresponding to a left-to-right depth-first tree
    traversal.

    Args:
        tree: a pytree to flatten.

    Returns:
        A pair where the first element is a list of leaf values and the second
        element is a treedef representing the structure of the flattened tree.

    See Also:
        - :func:`jax_scalify.tree.leaves`
        - :func:`jax_scalify.tree.structure`
        - :func:`jax_scalify.tree.unflatten`
    """
    return tree_util.tree_flatten(tree, is_leaf=is_scaled_leaf)


def leaves(
    tree: Any,
) -> list[Leaf]:
    """Gets the leaves (`Array` or `ScaledArray`) of a pytree.

    Args:
        tree: the pytree for which to get the leaves

    Returns:
        leaves: a list of tree leaves.

    See Also:
        - :func:`jax_scalify.tree.flatten`
        - :func:`jax_scalify.tree.structure`
        - :func:`jax_scalify.tree.unflatten`
    """
    return tree_util.tree_leaves(tree, is_leaf=is_scaled_leaf)


def map(f: Callable[..., Any], tree: Any, *rest: Any) -> Any:
    """Maps a multi-input function over pytree args to produce a new pytree.

    Args:
      f: function that takes ``1 + len(rest)`` arguments, to be applied at the
        corresponding leaves of the pytrees.
      tree: a pytree to be mapped over, with each leaf providing the first
        positional argument to ``f``.
      rest: a tuple of pytrees, each of which has the same structure as ``tree``
        or has ``tree`` as a prefix.

    Returns:
      A new pytree with the same structure as ``tree`` but with the value at each
      leaf given by ``f(x, *xs)`` where ``x`` is the value at the corresponding
      leaf in ``tree`` and ``xs`` is the tuple of values at corresponding nodes in
      ``rest``.

    See Also:
      - :func:`jax_scalify.tree.leaves`
      - :func:`jax_scalify.tree.reduce`
    """
    return tree_util.tree_map(f, tree, *rest, is_leaf=is_scaled_leaf)


def structure(tree: Any) -> tree_util.PyTreeDef:
    """Gets the treedef for a pytree, with `Array` or `ScaledArray` leaves.

    Args:
        tree: the pytree for which to get the leaves

    Returns:
        pytreedef: a PyTreeDef representing the structure of the tree.

    See Also:
      - :func:`jax_scalify.tree.flatten`
      - :func:`jax_scalify.tree.leaves`
      - :func:`jax_scalify.tree.unflatten`
    """
    return tree_util.tree_structure(tree, is_leaf=is_scaled_leaf)


# Alias of JAX tree unflatten.
unflatten = jax.tree_util.tree_unflatten



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from functools import partial

import chex
import jax
import jax.numpy as jnp
import ml_dtypes
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized
from numpy.typing import NDArray

from jax_scalify.core import scaled_array, scalify
from jax_scalify.ops import cast_on_backward, cast_on_forward, reduce_precision_on_forward


class ReducePrecisionDtypeTests(chex.TestCase):
    @parameterized.parameters(
        {"ml_dtype": ml_dtypes.float8_e4m3fn},
        {"ml_dtype": ml_dtypes.float8_e5m2},
    )
    def test__reduce_precision_on_forward__consistent_rounding_down(self, ml_dtype):
        # Values potentially "problematic" in FP8.
        values: NDArray[np.float16] = np.array([17, -17, 8, 1, 9, 11, 18], np.float16)
        out = reduce_precision_on_forward(values, dtype=ml_dtype)
        expected_out = values.astype(ml_dtype)
        assert out.dtype == values.dtype
        npt.assert_array_equal(out, expected_out)

    @parameterized.parameters(
        {"ml_dtype": ml_dtypes.float8_e4m3fn},
        {"ml_dtype": ml_dtypes.float8_e5m2},
    )
    def test__reduce_precision_on_forward__scalify_compatiblity(self, ml_dtype):
        values: NDArray[np.float16] = np.array([17, -17, 8, 1, 9, 11, 18], np.float16)
        arr = scaled_array(values, np.float32(1))
        out = scalify(partial(reduce_precision_on_forward, dtype=ml_dtype))(arr)

        npt.assert_array_equal(out.scale, arr.scale)
        npt.assert_array_equal(out, np.asarray(arr.data).astype(ml_dtype))


class CastOnForwardBackwardTests(chex.TestCase):
    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.parameters(
        {"dtype": jnp.float16},
        # TODO: uncomment when JAX 0.4+ used
        # {"dtype": jnp.float8_e4m3fn},
        # {"dtype": jnp.float8_e5m2},
    )
    def test__cast_on_forward_backward__proper_results(self, dtype):
        # Values potentially "problematic" in FP8.
        values: NDArray[np.float16] = np.array([17, -17, 8, 1, 9, 11, 18], np.float16)
        out_on_fwd = self.variant(partial(cast_on_forward, dtype=dtype))(values)
        out_on_bwd = self.variant(partial(cast_on_backward, dtype=dtype))(values)

        assert out_on_fwd.dtype == dtype
        assert out_on_bwd.dtype == values.dtype
        npt.assert_array_equal(out_on_fwd, jax.lax.convert_element_type(values, dtype))
        npt.assert_array_equal(out_on_bwd, values)

    @parameterized.parameters(
        {"dtype": jnp.float16},
        # TODO: uncomment when JAX 0.4+ used
        # {"dtype": jnp.float8_e4m3fn},
        # {"dtype": jnp.float8_e5m2},
    )
    def test__cast_on_backward__grad__proper_results(self, dtype):
        def fn(val, with_cast):
            if with_cast:
                val = cast_on_backward(val, dtype=dtype)
            val = val * val
            return jax.lax.reduce_sum_p.bind(val, axes=(0,))

        # Values potentially "problematic" in FP8.
        values: NDArray[np.float32] = np.array([17, -17, 8, 1, 9, 11, 18], np.float32)
        # Backward pass => gradient.
        grads = jax.grad(partial(fn, with_cast=True))(values)
        grads_ref = jax.grad(partial(fn, with_cast=False))(values)

        assert grads.dtype == dtype
        assert grads_ref.dtype == values.dtype
        npt.assert_array_equal(grads, jax.lax.convert_element_type(grads_ref, dtype))



# Copyright 2018 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified by Graphcore Ltd 2024.

"""A basic MNIST example using Numpy and JAX.

The primary aim here is simplicity and minimal dependencies.
"""


import time

import datasets
import jax.numpy as jnp
import ml_dtypes
import numpy as np
import numpy.random as npr
from jax import grad, jit, lax

import jax_scalify as jsa

# from functools import partial


def print_mean_std(name, v):
    """Debugging method/tool for JAX Scalify."""
    data, scale = jsa.lax.get_data_scale(v)
    # Always use np.float32, to avoid floating errors in descaling + stats.
    data = jsa.asarray(data, dtype=np.float32)
    m, s, min, max = np.mean(data), np.std(data), np.min(data), np.max(data)
    print(f"{name}: MEAN({m:.4f}) / STD({s:.4f}) / MIN({min:.4f}) / MAX({max:.4f}) / SCALE({scale:.4f})")


def logsumexp(a, axis=None, keepdims=False):
    dims = (axis,)
    amax = jnp.max(a, axis=dims, keepdims=keepdims)
    # FIXME: not proper scale propagation, introducing NaNs
    # amax = lax.stop_gradient(lax.select(jnp.isfinite(amax), amax, lax.full_like(amax, 0)))
    amax = lax.stop_gradient(amax)
    out = lax.sub(a, amax)
    out = lax.exp(out)
    out = lax.add(lax.log(jnp.sum(out, axis=dims, keepdims=keepdims)), amax)
    return out


def init_random_params(scale, layer_sizes, rng=npr.RandomState(0)):
    return [(scale * rng.randn(m, n), scale * rng.randn(n)) for m, n, in zip(layer_sizes[:-1], layer_sizes[1:])]


def predict(params, inputs, use_fp8=True):
    reduce_precision_on_forward = jsa.ops.reduce_precision_on_forward if use_fp8 else lambda x, d: x
    reduce_precision_on_backward = jsa.ops.reduce_precision_on_backward if use_fp8 else lambda x, d: x

    activations = inputs
    for w, b in params[:-1]:
        # Forward FP8 casting.
        w = reduce_precision_on_forward(w, ml_dtypes.float8_e4m3fn)
        activations = reduce_precision_on_forward(activations, ml_dtypes.float8_e4m3fn)
        # Matmul
        outputs = jnp.dot(activations, w)
        # Backward FP8 casting
        outputs = reduce_precision_on_backward(outputs, ml_dtypes.float8_e5m2)

        # Bias + relu
        outputs = outputs + b
        activations = jnp.maximum(outputs, 0)

    final_w, final_b = params[-1]
    # Forward FP8 casting.
    # final_w = jsa.ops.reduce_precision_on_forward(final_w, ml_dtypes.float8_e4m3fn)
    activations = reduce_precision_on_forward(activations, ml_dtypes.float8_e4m3fn)
    logits = jnp.dot(activations, final_w)
    # Backward FP8 casting
    logits = reduce_precision_on_backward(logits, ml_dtypes.float8_e5m2)

    logits = logits + final_b

    # Dynamic rescaling of the gradient, as logits gradient not properly scaled.
    logits = jsa.ops.dynamic_rescale_l2_grad(logits)
    logits = logits - logsumexp(logits, axis=1, keepdims=True)
    return logits


def loss(params, batch):
    inputs, targets = batch
    preds = predict(params, inputs)
    return -jnp.mean(jnp.sum(preds * targets, axis=1))


def accuracy(params, batch):
    inputs, targets = batch
    target_class = jnp.argmax(targets, axis=1)
    predicted_class = jnp.argmax(predict(params, inputs, use_fp8=False), axis=1)
    return jnp.mean(predicted_class == target_class)


if __name__ == "__main__":
    layer_sizes = [784, 512, 512, 10]
    param_scale = 0.1
    step_size = 0.1
    num_epochs = 10
    batch_size = 128

    training_dtype = np.float16
    scale_dtype = np.float32

    train_images, train_labels, test_images, test_labels = datasets.mnist()
    num_train = train_images.shape[0]
    num_complete_batches, leftover = divmod(num_train, batch_size)
    num_batches = num_complete_batches + bool(leftover)

    def data_stream():
        rng = npr.RandomState(0)
        while True:
            perm = rng.permutation(num_train)
            for i in range(num_batches):
                batch_idx = perm[i * batch_size : (i + 1) * batch_size]
                yield train_images[batch_idx], train_labels[batch_idx]

    batches = data_stream()
    params = init_random_params(param_scale, layer_sizes)
    # Transform parameters to `ScaledArray` and proper dtype.
    params = jsa.as_scaled_array(params, scale=scale_dtype(param_scale))
    params = jsa.tree.astype(params, training_dtype)

    @jit
    @jsa.scalify
    def update(params, batch):
        grads = grad(loss)(params, batch)
        return [(w - step_size * dw, b - step_size * db) for (w, b), (dw, db) in zip(params, grads)]

    for epoch in range(num_epochs):
        start_time = time.time()
        for _ in range(num_batches):
            batch = next(batches)
            # Scaled micro-batch + training dtype cast.
            batch = jsa.as_scaled_array(batch, scale=scale_dtype(1))
            batch = jsa.tree.astype(batch, training_dtype)

            with jsa.ScalifyConfig(rounding_mode=jsa.Pow2RoundMode.DOWN, scale_dtype=scale_dtype):
                params = update(params, batch)

        epoch_time = time.time() - start_time

        # Evaluation in float32, for consistency.
        raw_params = jsa.asarray(params, dtype=np.float32)
        train_acc = accuracy(raw_params, (train_images, train_labels))
        test_acc = accuracy(raw_params, (test_images, test_labels))
        print(f"Epoch {epoch} in {epoch_time:0.2f} sec")
        print(f"Training set accuracy {train_acc:0.5f}")
        print(f"Test set accuracy {test_acc:0.5f}")



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from typing import Any

import jax
import jax.numpy as jnp
import numpy as np

from .typing import Array


def safe_div(lhs: Array, rhs: Array) -> Array:
    """Safe (scalar) div: if rhs is zero, returns zero."""
    assert lhs.shape == ()
    assert rhs.shape == ()
    # assert lhs.dtype == rhs.dtype
    # Numpy inputs => direct computation.
    is_npy_inputs = isinstance(lhs, (np.number, np.ndarray)) and isinstance(rhs, (np.number, np.ndarray))
    if is_npy_inputs:
        return np.divide(lhs, rhs, out=np.array(0, dtype=rhs.dtype), where=rhs != 0)
    # JAX general implementation.
    return jax.lax.select(rhs == 0, rhs, jnp.divide(lhs, rhs))


def safe_reciprocal(val: Array) -> Array:
    """Safe (scalar) reciprocal: if val is zero, returns zero."""
    assert val.shape == ()
    # Numpy inputs => direct computation.
    if isinstance(val, (np.number, np.ndarray)):
        return np.reciprocal(val, out=np.array(0, dtype=val.dtype), where=val != 0)
    # JAX general implementation.
    return jax.lax.select(val == 0, val, jax.lax.reciprocal(val))


def python_scalar_as_numpy(val: Any) -> Any:
    """Convert Python scalar to Numpy scalar, if possible.

    Using by default JAX 32 bits precision, instead of 64 bits.

    Returning unchanged value if not any (bool, int, float).
    """
    if isinstance(val, bool):
        return np.bool_(val)
    elif isinstance(val, int):
        return np.int32(val)
    elif isinstance(val, float):
        return np.float32(val)
    return val



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from functools import partial

import chex
import jax.numpy as jnp
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized

from jax_scalify.core import Pow2RoundMode, pow2_decompose, pow2_round_down, pow2_round_up
from jax_scalify.core.pow2 import _exponent_bits_mask, get_mantissa


class Pow2DecomposePrimitveTests(chex.TestCase):
    @parameterized.parameters(
        {"dtype": np.float16},
        {"dtype": np.float32},
    )
    def test__exponent_bitmask__inf_value(self, dtype):
        val = _exponent_bits_mask[np.dtype(dtype)].view(dtype)
        expected_val = dtype(np.inf)
        npt.assert_equal(val, expected_val)

    @parameterized.product(
        val_exp=[
            (0, 0),
            (1, 1),
            (2.1, 2),
            (0.3, 0.25),
            (0.51, 0.5),
            (65500, 32768),
            # Test float16 sub-normals.
            (np.finfo(np.float16).smallest_normal, np.finfo(np.float16).smallest_normal),
            (np.finfo(np.float16).smallest_subnormal, np.finfo(np.float16).smallest_subnormal),
            (np.float16(3.123283386230469e-05), 3.0517578e-05),
        ],
        dtype=[np.float16, np.float32],
        scale_dtype=[np.float16, np.float32],
    )
    def test__pow2_decompose_round_down__numpy_implementation__proper_result(self, val_exp, dtype, scale_dtype):
        scale_dtype = np.float32
        vin, exp_scale = dtype(val_exp[0]), scale_dtype(val_exp[1])
        scale, vout = pow2_decompose(vin, scale_dtype, Pow2RoundMode.DOWN)

        assert isinstance(scale, (np.ndarray, np.number))
        assert isinstance(vout, (np.ndarray, np.number))
        assert scale.dtype == scale_dtype
        assert vout.dtype == vin.dtype
        # Always accurate when casting up to scale dtype.
        npt.assert_equal(scale * vout.astype(scale_dtype), vin.astype(scale_dtype))
        npt.assert_equal(scale, exp_scale)

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.product(
        val_exp=[
            (0, 0),
            (1, 1),
            (2.1, 2),
            (0.3, 0.25),
            (0.51, 0.5),
            (65500, 32768),
            # Test float16 sub-normals.
            (np.finfo(np.float16).smallest_normal, np.finfo(np.float16).smallest_normal),
            (np.finfo(np.float16).smallest_subnormal, np.finfo(np.float16).smallest_subnormal),
            (np.float16(3.123283386230469e-05), 3.0517578e-05),
            # Test float32 sub-normals: known bug!
            # (np.finfo(np.float32).smallest_normal, np.finfo(np.float32).smallest_normal),
            # (np.finfo(np.float32).smallest_subnormal, np.finfo(np.float32).smallest_subnormal),
        ],
        dtype=[np.float16, np.float32],
        scale_dtype=[np.float16, np.float32],
    )
    def test__pow2_decompose_round_down__jax_numpy__proper_result(self, val_exp, dtype, scale_dtype):
        vin, exp_scale = dtype(val_exp[0]), scale_dtype(val_exp[1])
        vin = jnp.array(vin)
        scale, vout = self.variant(lambda v: pow2_decompose(v, scale_dtype, Pow2RoundMode.DOWN))(vin)

        assert isinstance(scale, jnp.ndarray)
        assert isinstance(vout, jnp.ndarray)
        assert scale.dtype == scale_dtype
        assert vout.dtype == vin.dtype
        # Always accurate when casting up to scale dtype.
        npt.assert_equal(np.asarray(scale), exp_scale)
        npt.assert_equal(scale * np.array(vout, scale_dtype), np.asarray(vin, scale_dtype))

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.product(
        val_exp=[
            (+np.inf, np.inf, +np.inf),
            (-np.inf, np.inf, -np.inf),
            (np.nan, np.inf, np.nan),  # FIXME? scale == np.inf?
        ],
        dtype=[np.float16, np.float32],
        scale_dtype=[np.float16, np.float32],
    )
    def test__pow2_decompose_round_down__special_values(self, val_exp, dtype, scale_dtype):
        vin, exp_scale, exp_vout = dtype(val_exp[0]), scale_dtype(val_exp[1]), dtype(val_exp[2])
        scale, vout = self.variant(partial(pow2_decompose, scale_dtype=scale_dtype, mode=Pow2RoundMode.DOWN))(vin)
        npt.assert_equal(np.ravel(scale)[0], exp_scale)
        npt.assert_equal(np.ravel(vout)[0], exp_vout)

    @parameterized.product(
        val_exp=[(0, 0), (1, 1), (2.1, 2), (0.3, 0.25), (0.51, 0.5), (65500, 32768)],
        dtype=[np.float16, np.float32, np.float64],
    )
    def test__pow2_round_down__proper_rounding__multi_dtypes(self, val_exp, dtype):
        val, exp = dtype(val_exp[0]), dtype(val_exp[1])
        pow2_val = pow2_round_down(val)
        assert pow2_val.dtype == val.dtype
        assert pow2_val.shape == ()
        assert type(pow2_val) in {type(val), np.ndarray}
        npt.assert_equal(pow2_val, exp)

    @parameterized.product(
        val_exp=[(2.1, 4), (0.3, 0.5), (0.51, 1), (17000, 32768)],
        dtype=[np.float16],
    )
    def test__pow2_round_up__proper_rounding__multi_dtypes(self, val_exp, dtype):
        val, exp = dtype(val_exp[0]), dtype(val_exp[1])
        pow2_val = pow2_round_up(val)
        assert pow2_val.dtype == val.dtype
        assert type(pow2_val) in {type(val), np.ndarray}
        npt.assert_equal(pow2_val, exp)

    @parameterized.product(
        val_mant=[(1, 1), (2.1, 1.05), (0, 0), (0.51, 1.02), (65504, 1.9990234375)],
        dtype=[np.float16, np.float32],  # FIXME: float64 support in pure Numpy
    )
    def test__get_mantissa__proper_value__multi_dtypes(self, val_mant, dtype):
        val, mant = dtype(val_mant[0]), dtype(val_mant[1])
        val_mant = get_mantissa(val)
        assert val_mant.dtype == val.dtype
        assert val_mant.shape == ()
        assert type(val_mant) in {type(val), np.ndarray}
        npt.assert_equal(val_mant, mant)
        # Should be consistent with `pow2_round_down`. bitwise, not approximation.
        npt.assert_equal(mant * pow2_round_down(val), val)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from .base_scaling_primitives import (  # noqa: F401
    get_data_scale,
    get_data_scale_p,
    rebalance,
    set_scaling,
    set_scaling_p,
    stop_scaling,
    stop_scaling_p,
)
from .scaled_ops_common import *  # noqa: F401, F403
from .scaled_ops_l2 import *  # noqa: F401, F403



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized
from jax import lax

from jax_scalify.core import ScaledArray, find_registered_scaled_op, scaled_array
from jax_scalify.lax import scaled_div, scaled_dot_general, scaled_mul, scaled_reduce_window_sum


class ScaledTranslationDotPrimitivesTests(chex.TestCase):
    def setUp(self):
        super().setUp()
        # Use random state for reproducibility!
        self.rs = np.random.RandomState(42)

    @parameterized.parameters(
        {"ldtype": np.float32, "rdtype": np.float32},
        # {"ldtype": np.float32, "rdtype": np.float16}, # Not supported in JAX 0.3.x
        # {"ldtype": np.float16, "rdtype": np.float32},
        {"ldtype": np.float16, "rdtype": np.float16},
    )
    def test__scaled_dot_general__proper_scaling(self, ldtype, rdtype):
        # Reduction dimension: 5 => sqrt(5) ~ 2
        lhs = scaled_array(self.rs.rand(3, 5), 2.0, dtype=ldtype)
        rhs = scaled_array(self.rs.rand(5, 2), 4.0, dtype=rdtype)

        dimension_numbers = (((1,), (0,)), ((), ()))
        out = scaled_dot_general(lhs, rhs, dimension_numbers)
        expected_out = lax.dot_general(np.asarray(lhs), np.asarray(rhs), dimension_numbers)

        assert isinstance(out, ScaledArray)
        assert out.dtype == expected_out.dtype
        assert out.scale.dtype == np.float32  # TODO: more test coverage.
        npt.assert_almost_equal(out.scale, lhs.scale * rhs.scale * 2)
        npt.assert_array_almost_equal(out, expected_out, decimal=2)


class ScaledTranslationUnaryOpsTests(chex.TestCase):
    def setUp(self):
        super().setUp()
        # Use random state for reproducibility!
        self.rs = np.random.RandomState(42)

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.parameters(
        {"prim": lax.exp_p, "dtype": np.float16, "expected_scale": 1.0},  # FIXME!
        {"prim": lax.log_p, "dtype": np.float16, "expected_scale": 1.0},  # FIXME!
        {"prim": lax.neg_p, "dtype": np.float16, "expected_scale": 2.0},
        {"prim": lax.abs_p, "dtype": np.float16, "expected_scale": 2.0},
        {"prim": lax.cos_p, "dtype": np.float16, "expected_scale": 1.0},
        {"prim": lax.sin_p, "dtype": np.float16, "expected_scale": 1.0},
    )
    def test__scaled_unary_op__proper_result_and_scaling(self, prim, dtype, expected_scale):
        scaled_op, _ = find_registered_scaled_op(prim)
        val = scaled_array(self.rs.rand(3, 5), 2.0, dtype=dtype)
        out = self.variant(scaled_op)(val)
        expected_output = prim.bind(np.asarray(val))
        assert isinstance(out, ScaledArray)
        assert out.dtype == val.dtype
        assert out.scale.dtype == val.scale.dtype
        npt.assert_almost_equal(out.scale, expected_scale)
        # FIXME: higher precision for `log`?
        npt.assert_array_almost_equal(out, expected_output, decimal=3)

    def test__scaled_exp__large_scale_zero_values(self):
        scaled_op, _ = find_registered_scaled_op(lax.exp_p)
        # Scaled array, with values < 0 and scale overflowing in float16.
        val = scaled_array(np.array([0, -1, -2, -32768], np.float16), np.float32(32768 * 16))
        out = scaled_op(val)
        # Zero value should not be a NaN!
        npt.assert_array_almost_equal(out, [1, 0, 0, 0], decimal=2)

    def test__scaled_log__zero_large_values_large_scale(self):
        scaled_op, _ = find_registered_scaled_op(lax.log_p)
        # 0 + large values => proper log values, without NaN/overflow.
        val = scaled_array(np.array([0, 1], np.float16), np.float32(32768 * 16))
        out = scaled_op(val)
        # No NaN value + not overflowing!
        npt.assert_array_almost_equal(out, lax.log(val.to_array(np.float32)), decimal=2)


class ScaledTranslationBinaryOpsTests(chex.TestCase):
    def setUp(self):
        super().setUp()
        # Use random state for reproducibility!
        self.rs = np.random.RandomState(42)

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.product(
        prim=[lax.add_p, lax.sub_p, lax.mul_p, lax.div_p, lax.min_p, lax.max_p],
        dtype=[np.float16, np.float32],
        sdtype=[np.float16, np.float32],
    )
    def test__scaled_binary_op__proper_result_and_promotion(self, prim, dtype, sdtype):
        scaled_op, _ = find_registered_scaled_op(prim)
        # NOTE: direct construction to avoid weirdity between NumPy array and scalar!
        x = ScaledArray(np.array([-1.0, 2.0], dtype), sdtype(8.0))
        y = ScaledArray(np.array([1.5, 4.5], dtype), sdtype(2.0))
        # Ensure scale factor has the right dtype.
        assert x.scale.dtype == sdtype
        assert y.scale.dtype == sdtype

        z = self.variant(scaled_op)(x, y)
        expected_z = prim.bind(np.asarray(x), np.asarray(y))

        assert z.dtype == x.dtype
        assert z.scale.dtype == sdtype
        npt.assert_array_almost_equal(z, expected_z, decimal=4)

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.product(
        prim=[lax.add_p, lax.sub_p, lax.mul_p, lax.min_p, lax.max_p],
        dtype=[np.float16, np.float32],
        sdtype=[np.float16, np.float32],
    )
    def test__scaled_binary_op__proper_zero_scale_handling(self, prim, dtype, sdtype):
        scaled_op, _ = find_registered_scaled_op(prim)
        # NOTE: direct construction to avoid weirdity between NumPy array and scalar!
        x = ScaledArray(np.array([-1.0, 2.0], dtype), sdtype(0.0))
        y = ScaledArray(np.array([1.5, 4.5], dtype), sdtype(0.0))
        # Ensure scale factor has the right dtype.
        assert x.scale.dtype == sdtype
        assert y.scale.dtype == sdtype

        z = self.variant(scaled_op)(x, y)
        expected_z = prim.bind(np.asarray(x), np.asarray(y))

        assert z.dtype == x.dtype
        assert z.scale.dtype == sdtype
        npt.assert_array_almost_equal(z, expected_z, decimal=4)

    @parameterized.parameters(
        {"prim": lax.add_p},
        {"prim": lax.sub_p},
    )
    def test__scaled_addsub__proper_scaling(self, prim):
        scaled_op, _ = find_registered_scaled_op(prim)
        x = scaled_array([-1.0, 2.0], 4.0, dtype=np.float32)
        y = scaled_array([1.5, 4.5], 2.0, dtype=np.float32)
        z = scaled_op(x, y)
        assert isinstance(z, ScaledArray)
        assert z.dtype == x.dtype
        # Round down to power-of-2
        npt.assert_almost_equal(z.scale, 4)

    @parameterized.parameters(
        {"prim": lax.add_p},
        {"prim": lax.sub_p},
    )
    def test__scaled_addsub__not_overflowing_scale(self, prim):
        scaled_op, _ = find_registered_scaled_op(prim)
        x = scaled_array([-1.0, 2.0], np.float16(2.0), dtype=np.float16)
        y = scaled_array([1.5, 4.0], np.float16(1024.0), dtype=np.float16)
        z = scaled_op(x, y)
        assert z.scale.dtype == np.float16
        assert np.isfinite(z.scale)
        npt.assert_array_almost_equal(z, prim.bind(np.asarray(x, np.float32), np.asarray(y, np.float32)), decimal=6)

    @parameterized.product(
        prim=[lax.min_p, lax.max_p],
    )
    def test__scaled_minmax__static_zero_scale_propagation(self, prim):
        scaled_op, _ = find_registered_scaled_op(prim)
        x = scaled_array([-1.0, 2.0], 4.0, dtype=np.float32)
        y = scaled_array([1.5, 4.5], 0.0, dtype=np.float32)
        z = scaled_op(x, y)
        assert isinstance(z, ScaledArray)
        assert z.dtype == x.dtype
        # Keep the lhs scale.
        npt.assert_almost_equal(z.scale, 4.0)

    @parameterized.product(
        prim=[lax.min_p, lax.max_p],
    )
    def test__scaled_minmax__static_inf_scale_propagation(self, prim):
        scaled_op, _ = find_registered_scaled_op(prim)
        x = scaled_array([-1.0, 2.0], 4.0, dtype=np.float32, npapi=np)
        y = scaled_array([-np.inf, np.inf], np.inf, dtype=np.float32, npapi=np)
        z = scaled_op(x, y)
        assert isinstance(z, ScaledArray)
        assert z.dtype == x.dtype
        # Keep the lhs scale.
        npt.assert_almost_equal(z.scale, 4.0)

    def test__scaled_mul__proper_scaling(self):
        x = scaled_array([-2.0, 2.0], 3, dtype=np.float32)
        y = scaled_array([1.5, 1.5], 2, dtype=np.float32)
        z = scaled_mul(x, y)
        assert isinstance(z, ScaledArray)
        assert z.scale == 6
        npt.assert_array_almost_equal(z, np.asarray(x) * np.asarray(y))

    def test__scaled_div__proper_scaling(self):
        x = scaled_array([-2.0, 2.0], 3.0, dtype=np.float32)
        y = scaled_array([1.5, 1.5], 2.0, dtype=np.float32)
        z = scaled_div(x, y)
        assert isinstance(z, ScaledArray)
        assert z.scale == 1.5
        npt.assert_array_almost_equal(z, np.asarray(x) / np.asarray(y))


class ScaledTranslationReducePrimitivesTests(chex.TestCase):
    def setUp(self):
        super().setUp()
        # Use random state for reproducibility!
        self.rs = np.random.RandomState(42)

    @parameterized.parameters(
        {"reduce_prim": lax.reduce_sum_p, "expected_scale": 2 * 2},
        {"reduce_prim": lax.reduce_prod_p, "expected_scale": 2**5},
        {"reduce_prim": lax.reduce_min_p, "expected_scale": 2},
        {"reduce_prim": lax.reduce_max_p, "expected_scale": 2},
    )
    def test__scaled_reduce__single_axis__proper_scaling(self, reduce_prim, expected_scale):
        axes = (0,)
        # NOTE: float16 useful for checking dtype promotion!
        val = scaled_array(self.rs.rand(5), 2.0, dtype=np.float16)
        scaled_reduce_op, _ = find_registered_scaled_op(reduce_prim)
        out = scaled_reduce_op(val, axes=axes)

        assert isinstance(out, ScaledArray)
        assert out.shape == ()
        assert out.dtype == val.dtype
        npt.assert_almost_equal(out.scale, expected_scale)
        npt.assert_array_almost_equal(out, reduce_prim.bind(np.asarray(val), axes=axes))

    def test__scaled_reduce_window_sum__proper_result(self):
        val = scaled_array(self.rs.rand(5), 2.0, dtype=np.float32)
        out = scaled_reduce_window_sum(
            val,
            window_dimensions=(3,),
            window_strides=(1,),
            padding=((1, 0),),
            base_dilation=(1,),
            window_dilation=(1,),
        )
        assert isinstance(out, ScaledArray)
        assert out.shape == (4,)
        assert out.dtype == val.dtype
        npt.assert_almost_equal(out.scale, val.scale)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import jax.numpy as jnp
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized

from jax_scalify.core.utils import Array, python_scalar_as_numpy, safe_div, safe_reciprocal


class SafeDivOpTests(chex.TestCase):
    @parameterized.parameters(
        {"lhs": np.float16(0), "rhs": np.float16(0)},
        {"lhs": np.float32(0), "rhs": np.float32(0)},
        {"lhs": np.float16(2), "rhs": np.float16(0)},
        {"lhs": np.float32(4), "rhs": np.float32(0)},
    )
    def test__safe_div__zero_div__numpy_inputs(self, lhs, rhs):
        out = safe_div(lhs, rhs)
        assert isinstance(out, (np.number, np.ndarray))
        assert out.dtype == lhs.dtype
        npt.assert_equal(out, 0)

    @parameterized.parameters(
        {"lhs": np.float16(0), "rhs": jnp.float16(0)},
        {"lhs": jnp.float32(0), "rhs": np.float32(0)},
        {"lhs": jnp.float16(2), "rhs": np.float16(0)},
        {"lhs": np.float32(4), "rhs": jnp.float32(0)},
    )
    def test__safe_div__zero_div__jax_inputs(self, lhs, rhs):
        out = safe_div(lhs, rhs)
        assert isinstance(out, Array)
        assert out.dtype == lhs.dtype
        npt.assert_almost_equal(out, 0)

    @parameterized.parameters(
        {"val": np.float16(0)},
        {"val": jnp.float16(0)},
    )
    def test__safe_reciprocal__zero_div(self, val):
        out = safe_reciprocal(val)
        assert out.dtype == val.dtype
        npt.assert_almost_equal(out, 0)


def test__python_scalar_as_numpy__proper_convertion():
    npt.assert_equal(python_scalar_as_numpy(False), np.bool_(False))
    npt.assert_equal(python_scalar_as_numpy(4), np.int32(4))
    npt.assert_equal(python_scalar_as_numpy(3.2), np.float32(3.2))



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from typing import Any, Optional, Sequence

import jax
import jax.core
import jax.numpy as jnp
import numpy as np
from jax import lax

from jax_scalify import core
from jax_scalify.core import (
    Array,
    DTypeLike,
    ScaledArray,
    Shape,
    as_scaled_array,
    get_scale_dtype,
    is_static_anyscale,
    is_static_zero,
    safe_div,
)

from .base_scaling_primitives import scaled_set_scaling


def _get_data(val: Any) -> Array:
    if isinstance(val, ScaledArray):
        return val.data
    return val


def check_scalar_scales(*args: ScaledArray) -> Array:
    """Check all ScaledArrays have scalar scaling."""
    for val in args:
        assert np.ndim(val.scale) == 0


def promote_types(*args: DTypeLike) -> DTypeLike:
    """Find a common promotion dtype."""
    outdtype = args[0]
    for val in args[1:]:
        outdtype = jnp.promote_types(outdtype, val)
    return outdtype


def promote_scale_types(*args: ScaledArray) -> Sequence[ScaledArray]:
    """Promote scale datatypes to a common one.

    Note: we are using JAX Numpy promotion, to avoid 64bits types by default.
    """
    if len(args) == 1:
        return args
    # Find a common scale datatype.
    scale_dtype = args[0].scale.dtype
    for val in args[1:]:
        scale_dtype = jnp.promote_types(scale_dtype, val.scale.dtype)

    outputs = [ScaledArray(v.data, v.scale.astype(scale_dtype)) for v in args]
    return outputs


@core.register_scaled_lax_op
def scaled_stop_gradient(val: ScaledArray) -> ScaledArray:
    # Stop gradients on both data and scale tensors.
    return ScaledArray(lax.stop_gradient(val.data), lax.stop_gradient(val.scale))


@core.register_scaled_lax_op
def scaled_reshape(A: ScaledArray, new_sizes: Sequence[int], dimensions: Optional[Sequence[int]]) -> ScaledArray:
    return ScaledArray(lax.reshape(A.data, new_sizes=new_sizes, dimensions=dimensions), A.scale)


@core.register_scaled_lax_op
def scaled_broadcast_in_dim(A: ScaledArray, shape: Shape, broadcast_dimensions: Sequence[int]) -> ScaledArray:
    return ScaledArray(lax.broadcast_in_dim(A.data, shape=shape, broadcast_dimensions=broadcast_dimensions), A.scale)


@core.register_scaled_lax_op
def scaled_convert_element_type(A: ScaledArray, new_dtype: DTypeLike, weak_type: bool = False) -> ScaledArray:
    # NOTE: by default, no rescaling done before casting.
    # Choice of adding an optional rescaling op before is up to the user (and which strategy to use).
    # NOTE bis: scale not casted as well by default!
    return ScaledArray(lax.convert_element_type(A.data, new_dtype=new_dtype), A.scale)


@core.register_scaled_lax_op
def scaled_reduce_precision(A: ScaledArray, exponent_bits: int, mantissa_bits: int) -> ScaledArray:
    # Applying precision reduction only data term.
    return ScaledArray(lax.reduce_precision(A.data, exponent_bits=exponent_bits, mantissa_bits=mantissa_bits), A.scale)


@core.register_scaled_lax_op
def scaled_concatenate(operands: Sequence[ScaledArray], dimension: int) -> ScaledArray:
    # TODO: inputs checking (dtype and cie).
    scales = jnp.array([v.scale for v in operands])
    # Max rescaling of the collection of operands. Preserving pow2 scaling.
    # TODO: explore alternative strategies?
    outdtype = operands[0].dtype
    scale_max = jnp.max(scales)
    datas = [v.data * safe_div(v.scale, scale_max).astype(outdtype) for v in operands]
    data_concat = lax.concatenate(datas, dimension=dimension)
    return ScaledArray(data_concat, scale_max)


@core.register_scaled_lax_op
def scaled_slice(
    A: ScaledArray, start_indices: Sequence[int], limit_indices: Sequence[int], strides: Optional[Sequence[int]] = None
) -> ScaledArray:
    return ScaledArray(
        lax.slice(A.data, start_indices=start_indices, limit_indices=limit_indices, strides=strides), A.scale
    )


@core.register_scaled_lax_op
def scaled_transpose(A: ScaledArray, permutation: Sequence[int]) -> ScaledArray:
    return ScaledArray(lax.transpose(A.data, permutation=permutation), A.scale)


@core.register_scaled_lax_op
def scaled_rev(val: ScaledArray, dimensions: Sequence[int]) -> ScaledArray:
    return ScaledArray(lax.rev(val.data, dimensions=dimensions), val.scale)


@core.register_scaled_lax_op
def scaled_pad(val: ScaledArray, padding_value: Any, padding_config: Any) -> ScaledArray:
    # Only supporting constant zero padding for now.
    assert np.all(is_static_zero(padding_value))
    return ScaledArray(lax.pad(val.data, np.array(0, val.dtype), padding_config), val.scale)


@core.register_scaled_lax_op
def scaled_argmax(val: ScaledArray, axes: Sequence[int], index_dtype: DTypeLike) -> Array:
    # Note: returning a normal `int` Array.
    return lax.argmax_p.bind(val.data, axes=axes, index_dtype=index_dtype)


@core.register_scaled_lax_op
def scaled_argmin(val: ScaledArray, axes: Sequence[int], index_dtype: DTypeLike) -> Array:
    # Note: returning a normal `int` Array.
    return lax.argmin_p.bind(val.data, axes=axes, index_dtype=index_dtype)


@core.register_scaled_lax_op
def scaled_neg(val: ScaledArray) -> ScaledArray:
    return ScaledArray(-val.data, val.scale)


@core.register_scaled_lax_op
def scaled_abs(val: ScaledArray) -> ScaledArray:
    return ScaledArray(jax.lax.abs(val.data), val.scale)


@core.register_scaled_lax_op
def scaled_mul(lhs: ScaledArray, rhs: ScaledArray) -> ScaledArray:
    # TODO: understand when promotion is really required?
    lhs, rhs = as_scaled_array((lhs, rhs))  # type:ignore
    return ScaledArray(lhs.data * rhs.data, lhs.scale * rhs.scale)


@core.register_scaled_lax_op
def scaled_div(lhs: ScaledArray, rhs: ScaledArray) -> ScaledArray:
    # TODO: understand when promotion is really required?
    # lhs, rhs = as_scaled_array((lhs, rhs))  # type:ignore
    # TODO: investigate different rule?
    return ScaledArray(lhs.data / rhs.data, lhs.scale / rhs.scale)


@core.register_scaled_lax_op
def scaled_sign(val: ScaledArray) -> Array:
    assert isinstance(val, ScaledArray)
    # Just need to run `lax.sign` on main data.
    return ScaledArray(lax.sign(val.data), np.array(1, dtype=val.scale.dtype))


@core.register_scaled_lax_op
def scaled_is_finite(val: ScaledArray) -> Array:
    assert isinstance(val, ScaledArray)
    if np.issubdtype(val.scale.dtype, np.integer):
        # Integer scale case => only check the data component.
        return lax.is_finite(val.data)
    # Both data & scale need to be finite!
    return lax.and_p.bind(lax.is_finite(val.data), lax.is_finite(val.scale))


def scaled_boolean_binary_op(lhs: ScaledArray, rhs: ScaledArray, prim: jax.core.Primitive) -> Array:
    """Generic implementation of any boolean binary operation."""
    assert isinstance(lhs, ScaledArray)
    assert isinstance(rhs, ScaledArray)
    # FIXME: fix this absolute horror!
    # TODO: use max scale + special case for scalars.
    return prim.bind(lhs.to_array(dtype=np.float32), rhs.to_array(dtype=np.float32))


@core.register_scaled_lax_op
def scaled_eq(lhs: ScaledArray, rhs: ScaledArray) -> Array:
    return scaled_boolean_binary_op(lhs, rhs, lax.eq_p)


@core.register_scaled_lax_op
def scaled_ne(lhs: ScaledArray, rhs: ScaledArray) -> Array:
    return scaled_boolean_binary_op(lhs, rhs, lax.ne_p)


@core.register_scaled_lax_op
def scaled_gt(lhs: ScaledArray, rhs: ScaledArray) -> Array:
    return scaled_boolean_binary_op(lhs, rhs, lax.gt_p)


@core.register_scaled_lax_op
def scaled_ge(lhs: ScaledArray, rhs: ScaledArray) -> Array:
    return scaled_boolean_binary_op(lhs, rhs, lax.ge_p)


@core.register_scaled_lax_op
def scaled_lt(lhs: ScaledArray, rhs: ScaledArray) -> Array:
    return scaled_boolean_binary_op(lhs, rhs, lax.lt_p)


@core.register_scaled_lax_op
def scaled_le(lhs: ScaledArray, rhs: ScaledArray) -> Array:
    return scaled_boolean_binary_op(lhs, rhs, lax.le_p)


def scaled_minmax(prim: jax.core.Primitive, lhs: ScaledArray, rhs: ScaledArray) -> ScaledArray:
    """General min/max scaled translation: propagating the largest input scale."""
    check_scalar_scales(lhs, rhs)
    # Specific rule if lhs/rhs is zero or inf => propagate the other term scale.
    if np.all(is_static_anyscale(lhs)):
        return ScaledArray(prim.bind(lhs.data, rhs.data), rhs.scale)
    if np.all(is_static_anyscale(rhs)):
        return ScaledArray(prim.bind(lhs.data, rhs.data), lhs.scale)

    # Power-of-2 stable!
    output_scale = lax.max(lhs.scale, rhs.scale)
    # TODO: isolate this "binary" rescale logic into separate function.
    outdtype = jnp.promote_types(lhs.dtype, rhs.dtype)
    lhs_rescale = safe_div(lhs.scale, output_scale).astype(outdtype)
    rhs_rescale = safe_div(rhs.scale, output_scale).astype(outdtype)
    output_data = prim.bind(lhs_rescale * lhs.data, rhs_rescale * rhs.data)
    return ScaledArray(output_data, output_scale)


@core.register_scaled_lax_op
def scaled_min(lhs: ScaledArray, rhs: ScaledArray) -> ScaledArray:
    return scaled_minmax(lax.min_p, lhs, rhs)


@core.register_scaled_lax_op
def scaled_max(lhs: ScaledArray, rhs: ScaledArray) -> ScaledArray:
    return scaled_minmax(lax.max_p, lhs, rhs)


##################################################################
# Default scaled ops implementation #
##################################################################
def scaled_op_default_translation(
    prim: jax.core.Primitive, args: Sequence[ScaledArray], outscale: Optional[Array] = None
) -> ScaledArray:
    """Scaled op default translation of a JAX primitive: unscaling inputs + calling normal primitive.

    Args:
        prim: JAX primitive
        args: Input arguments.
        outscale: Output scale to use.
    """
    inputs = [core.asarray(v) for v in args]
    output = prim.bind(*inputs)
    # Rescale output, if necessary.
    if outscale is None:
        output_scale_dtype = promote_types(*[v.scale.dtype for v in args])
        return ScaledArray(output, np.array(1.0, dtype=output_scale_dtype))
    output_scaled = scaled_set_scaling(output, outscale)
    return output_scaled


@core.register_scaled_lax_op
def scaled_exp(val: ScaledArray) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    # Estimate in FP32, to avoid NaN when "descaling" the array.
    # Otherwise: issues for representing properly 0 and +-Inf.
    arr = val.to_array(dtype=np.float32).astype(val.dtype)
    scale = np.array(1, dtype=val.scale.dtype)
    return ScaledArray(lax.exp(arr), scale)


@core.register_scaled_lax_op
def scaled_log(val: ScaledArray) -> ScaledArray:
    assert isinstance(val, ScaledArray)
    # Log of data & scale components.
    log_data = lax.log(val.data)
    log_scale = lax.log(val.scale).astype(val.dtype)
    data = log_data + log_scale
    scale = np.array(1, dtype=val.scale.dtype)
    return ScaledArray(data, scale)


@core.register_scaled_lax_op
def scaled_select_n(which: Array, *cases: ScaledArray) -> ScaledArray:
    outscale_dtype = promote_types(*[get_scale_dtype(v) for v in cases])
    # Get the max scale for renormalizing.
    # TODO: use `get_data_scale` primitive.
    scale_cases = [v.scale if isinstance(v, ScaledArray) else np.array(1, outscale_dtype) for v in cases]
    scales_arr = jnp.array(scale_cases)
    outscale = jnp.max(scales_arr)
    # `data` components, renormalized.
    data_cases = [_get_data(v) * (s / outscale).astype(v.dtype) for s, v in zip(scale_cases, cases)]
    # Apply normal select!
    outdata = lax.select_n(which, *data_cases)
    return ScaledArray(outdata, outscale)


@core.register_scaled_lax_op
def scaled_cos(val: ScaledArray) -> ScaledArray:
    return scaled_op_default_translation(lax.cos_p, [val])


@core.register_scaled_lax_op
def scaled_sin(val: ScaledArray) -> ScaledArray:
    return scaled_op_default_translation(lax.sin_p, [val])


@core.register_scaled_lax_op
def scaled_integer_pow(A: ScaledArray, y: int) -> ScaledArray:
    output_scale = lax.integer_pow(A.scale, y)
    output_data = lax.integer_pow(A.data, y)
    return ScaledArray(output_data, output_scale)


@core.register_scaled_lax_op
def scaled_sqrt(val: ScaledArray) -> ScaledArray:
    output_scale = lax.sqrt(val.scale)
    output_data = lax.sqrt(val.data)
    return ScaledArray(output_data, output_scale)



# Copyright 2018 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified by Graphcore Ltd 2024.

"""A basic MNIST example using Numpy and JAX.

The primary aim here is simplicity and minimal dependencies.
"""


import time

import datasets
import jax
import jax.numpy as jnp
import numpy as np
import numpy.random as npr
from jax import grad, jit, lax

import jax_scalify as jsa

# from jax.scipy.special import logsumexp


def logsumexp(a, axis=None, keepdims=False):
    dims = (axis,)
    amax = jnp.max(a, axis=dims, keepdims=keepdims)
    # FIXME: not proper scale propagation, introducing NaNs
    # amax = lax.stop_gradient(lax.select(jnp.isfinite(amax), amax, lax.full_like(amax, 0)))
    amax = lax.stop_gradient(amax)
    out = lax.sub(a, amax)
    out = lax.exp(out)
    out = lax.add(lax.log(jnp.sum(out, axis=dims, keepdims=keepdims)), amax)
    return out


def init_random_params(scale, layer_sizes, rng=npr.RandomState(0)):
    return [(scale * rng.randn(m, n), scale * rng.randn(n)) for m, n, in zip(layer_sizes[:-1], layer_sizes[1:])]


def predict(params, inputs):
    activations = inputs
    for w, b in params[:-1]:
        # Matmul + relu
        outputs = jnp.dot(activations, w) + b
        activations = jax.nn.relu(outputs)

    final_w, final_b = params[-1]
    logits = jnp.dot(activations, final_w) + final_b
    # Dynamic rescaling of the gradient, as logits gradient not properly scaled.
    # logits = jsa.ops.dynamic_rescale_l2_grad(logits)
    logits = logits - logsumexp(logits, axis=1, keepdims=True)
    return logits


def loss(params, batch):
    inputs, targets = batch
    preds = predict(params, inputs)
    targets = jsa.lax.rebalance(targets, np.float32(1 / 8))
    return -jnp.mean(jnp.sum(preds * targets, axis=1))


def accuracy(params, batch):
    inputs, targets = batch
    target_class = jnp.argmax(targets, axis=1)
    predicted_class = jnp.argmax(predict(params, inputs), axis=1)
    return jnp.mean(predicted_class == target_class)


if __name__ == "__main__":
    layer_sizes = [784, 512, 512, 10]
    param_scale = 0.1
    step_size = 0.1
    num_epochs = 10
    batch_size = 128

    training_dtype = np.float16
    scale_dtype = np.float32

    train_images, train_labels, test_images, test_labels = datasets.mnist()
    num_train = train_images.shape[0]
    num_complete_batches, leftover = divmod(num_train, batch_size)
    num_batches = num_complete_batches + bool(leftover)

    def data_stream():
        rng = npr.RandomState(0)
        while True:
            perm = rng.permutation(num_train)
            for i in range(num_batches):
                batch_idx = perm[i * batch_size : (i + 1) * batch_size]
                yield train_images[batch_idx], train_labels[batch_idx]

    batches = data_stream()
    params = init_random_params(param_scale, layer_sizes)
    # Transform parameters to `ScaledArray` and proper dtype.
    params = jsa.as_scaled_array(params, scale=scale_dtype(param_scale))
    params = jsa.tree.astype(params, training_dtype)

    @jit
    @jsa.scalify
    def update(params, batch):
        grads = grad(loss)(params, batch)
        return [(w - step_size * dw, b - step_size * db) for (w, b), (dw, db) in zip(params, grads)]

    for epoch in range(num_epochs):
        start_time = time.time()
        for _ in range(num_batches):
            batch = next(batches)
            # Scaled micro-batch + training dtype cast.
            batch = jsa.as_scaled_array(batch, scale=scale_dtype(1))
            batch = jsa.tree.astype(batch, training_dtype)

            with jsa.ScalifyConfig(rounding_mode=jsa.Pow2RoundMode.DOWN, scale_dtype=scale_dtype):
                params = update(params, batch)

        epoch_time = time.time() - start_time

        # Evaluation in normal/unscaled float32, for consistency.
        raw_params = jsa.asarray(params, dtype=np.float32)
        train_acc = accuracy(raw_params, (train_images, train_labels))
        test_acc = accuracy(raw_params, (test_images, test_labels))
        print(f"Epoch {epoch} in {epoch_time:0.2f} sec")
        print(f"Training set accuracy {train_acc:0.5f}")
        print(f"Test set accuracy {test_acc:0.5f}")



# Copyright 2018 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified by Graphcore Ltd 2024.


"""Datasets used in examples."""


import array
import gzip
import os
import pickle
import struct
import tarfile
import urllib.request
from os import path

import numpy as np

_DATA = "/tmp/jax_example_data/"


def _download(url, filename):
    """Download a url to a file in the JAX data temp directory."""
    if not path.exists(_DATA):
        os.makedirs(_DATA)
    out_file = path.join(_DATA, filename)
    if not path.isfile(out_file):
        urllib.request.urlretrieve(url, out_file)
        print(f"downloaded {url} to {_DATA}")


def _partial_flatten(x):
    """Flatten all but the first dimension of an ndarray."""
    return np.reshape(x, (x.shape[0], -1))


def _one_hot(x, k, dtype=np.float32):
    """Create a one-hot encoding of x of size k."""
    return np.array(x[:, None] == np.arange(k), dtype)


def _unzip(file):
    file = tarfile.open(file)
    file.extractall(_DATA)
    file.close()
    return


def _unpickle(file):
    with open(file, "rb") as fo:
        dict = pickle.load(fo, encoding="bytes")
    return dict


def mnist_raw():
    """Download and parse the raw MNIST dataset."""
    # CVDF mirror of http://yann.lecun.com/exdb/mnist/
    base_url = "https://storage.googleapis.com/cvdf-datasets/mnist/"

    def parse_labels(filename):
        with gzip.open(filename, "rb") as fh:
            _ = struct.unpack(">II", fh.read(8))
            return np.array(array.array("B", fh.read()), dtype=np.uint8)

    def parse_images(filename):
        with gzip.open(filename, "rb") as fh:
            _, num_data, rows, cols = struct.unpack(">IIII", fh.read(16))
            return np.array(array.array("B", fh.read()), dtype=np.uint8).reshape(num_data, rows, cols)

    for filename in [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
    ]:
        _download(base_url + filename, filename)

    train_images = parse_images(path.join(_DATA, "train-images-idx3-ubyte.gz"))
    train_labels = parse_labels(path.join(_DATA, "train-labels-idx1-ubyte.gz"))
    test_images = parse_images(path.join(_DATA, "t10k-images-idx3-ubyte.gz"))
    test_labels = parse_labels(path.join(_DATA, "t10k-labels-idx1-ubyte.gz"))

    return train_images, train_labels, test_images, test_labels


def mnist(permute_train=False):
    """Download, parse and process MNIST data to unit scale and one-hot labels."""
    train_images, train_labels, test_images, test_labels = mnist_raw()

    train_images = _partial_flatten(train_images) / np.float32(255.0)
    test_images = _partial_flatten(test_images) / np.float32(255.0)
    train_labels = _one_hot(train_labels, 10)
    test_labels = _one_hot(test_labels, 10)

    if permute_train:
        perm = np.random.RandomState(0).permutation(train_images.shape[0])
        train_images = train_images[perm]
        train_labels = train_labels[perm]

    return train_images, train_labels, test_images, test_labels


def cifar_raw():
    """Download, unzip and parse the raw cifar dataset."""

    filename = "cifar-10-python.tar.gz"
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    _download(url, filename)
    _unzip(path.join(_DATA, filename))

    data_batches = ["data_batch_1", "data_batch_2", "data_batch_3", "data_batch_4", "data_batch_5"]
    data = []
    labels = []
    for batch in data_batches:
        tmp_dict = _unpickle(path.join(_DATA, "cifar-10-batches-py", batch))
        data.append(tmp_dict[b"data"])
        labels.append(tmp_dict[b"labels"])
    train_images = np.concatenate(data)
    train_labels = np.concatenate(labels)

    test_dict = _unpickle(path.join(_DATA, "cifar-10-batches-py", "test_batch"))
    test_images = test_dict[b"data"]
    test_labels = np.array(test_dict[b"labels"])

    return train_images, train_labels, test_images, test_labels


def cifar(permute_train=False):
    """Download, parse and process cifar data to unit scale and one-hot labels."""

    train_images, train_labels, test_images, test_labels = cifar_raw()

    train_images = train_images / np.float32(255.0)
    test_images = test_images / np.float32(255.0)
    train_labels = _one_hot(train_labels, 10)
    test_labels = _one_hot(test_labels, 10)

    if permute_train:
        perm = np.random.RandomState(0).permutation(train_images.shape[0])
        train_images = train_images[perm]
        train_labels = train_labels[perm]

    return train_images, train_labels, test_images, test_labels



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import logging
from typing import Any, Dict, Optional, Sequence, Union

import numpy as np
from jax import core
from jax.interpreters import mlir
from jax.interpreters.mlir import LoweringRuleContext, ir, ir_constant

from jax_scalify.core import (
    Array,
    DTypeLike,
    ScaledArray,
    ScaledPrimitiveType,
    asarray,
    get_scalify_config,
    is_static_one_scalar,
    register_scaled_op,
    safe_div,
    safe_reciprocal,
)

set_scaling_p = core.Primitive("set_scaling_p")
"""`set_scaling` JAX primitive.

In standard JAX, this is just an identity operation, ignoring the `scale`
input, just returning unchanged the `data` component.

In JAX Scalify mode, it will rebalance the data term to
return a ScaledArray semantically equivalent.

NOTE: there is specific corner case of passing zero to `set_scaling`. In this
situation, the tensor is assumed to be zeroed by the user.
"""


def set_scaling(values: Array, scale: Array) -> Array:
    """`set_scaling` primitive call method."""
    return set_scaling_p.bind(values, scale)


def set_scaling_impl(values: Array, scale: Array) -> Array:
    """Set scaling general implementation.

    Need to work on the following inputs combinations:
        - (Array, Array) -> Array
        - (ScaledArray, Array) -> ScaledArray
        - (ScaledArray, ScaledArray) -> ScaledArray
    with Numpy or JAX arrays (and keep the proper type for output).
    """
    assert scale.shape == ()
    if isinstance(values, ScaledArray):
        # Automatic promotion should ensure we always get a scaled scalar here!
        scale_value = asarray(scale)
        # Rebalancing data tensor using the new scale.
        data = values.data * safe_div(values.scale, scale_value).astype(values.dtype)
        return ScaledArray(data, scale_value)
    # No scaled array => no-op.
    return values


def set_scaling_abstract_eval(values: core.ShapedArray, scale: core.ShapedArray) -> core.ShapedArray:
    return values


def set_scaling_mlir_lowering(
    ctx: LoweringRuleContext, *args: Union[ir.Value, Sequence[ir.Value]]
) -> Sequence[Union[ir.Value, Sequence[ir.Value]]]:
    # Just forwarding `values` term, ignoring the `scale`.
    return (args[0],)


def scaled_set_scaling(values: ScaledArray, scale: ScaledArray) -> ScaledArray:
    """Scaled `set_scaling` implementation: rebalancing the data using the new scale value."""
    # Trivial case of scale == 1
    if is_static_one_scalar(scale):
        if isinstance(values, ScaledArray):
            return values
        return ScaledArray(values, scale)
    assert scale.shape == ()
    # Automatic promotion should ensure we always get a scaled scalar here!
    scale_value = asarray(scale)
    if not isinstance(values, ScaledArray):
        # Simple case, with no pre-existing scale.
        return ScaledArray(values * safe_reciprocal(scale_value.astype(values.dtype)), scale_value)
    # Rebalancing data tensor using the new scale.
    data = values.data * safe_div(values.scale, scale_value).astype(values.dtype)
    return ScaledArray(data, scale_value)


# Register as standard JAX primitive
set_scaling_p.multiple_results = False
set_scaling_p.def_abstract_eval(set_scaling_abstract_eval)
set_scaling_p.def_impl(set_scaling_impl)
mlir.register_lowering(set_scaling_p, set_scaling_mlir_lowering)
# Register "scaled" translation.
register_scaled_op(set_scaling_p, scaled_set_scaling, ScaledPrimitiveType.ALWAYS_SCALE)


stop_scaling_p = core.Primitive("stop_scaling_p")
"""`stop_scaling` JAX primitive.

In standard JAX, this is just an identity operation (with optional casting).

In JAX Scalify mode, it will return the value tensor, with optional casting.

Similar in principle to `jax.lax.stop_gradient`
"""


def stop_scaling(values: Array, dtype: Optional[DTypeLike] = None) -> Array:
    """`stop_scaling` primitive call method."""
    return stop_scaling_p.bind(values, dtype=dtype)


def stop_scaling_impl(values: Array, dtype: Optional[DTypeLike]) -> Array:
    if isinstance(values, ScaledArray):
        return values.to_array(dtype=dtype)
    if dtype is not None:
        values = values.astype(dtype)
    return values


def stop_scaling_abstract_eval(values: core.ShapedArray, dtype: Optional[DTypeLike]) -> core.ShapedArray:
    return values.update(dtype=dtype)


def stop_scaling_mlir_lowering(
    ctx: LoweringRuleContext, *args: Union[ir.Value, Sequence[ir.Value]], **params: Dict[str, Any]
) -> Sequence[Union[ir.Value, Sequence[ir.Value]]]:
    dtype = params.get("dtype", None)
    if dtype is not None:
        # TODO: caching of the MLIR lowered function?
        stop_scaling_mlir_fn = mlir.lower_fun(lambda x: x.astype(dtype), multiple_results=False)
        return stop_scaling_mlir_fn(ctx, *args)
    # By default: forward tensor.
    return (args[0],)


def scaled_stop_scaling(values: ScaledArray, dtype: Optional[DTypeLike] = None) -> Array:
    """Scaled `stop_scaling` implementation: returning tensor values (with optional cast)."""
    assert isinstance(values, ScaledArray)
    # TODO/FIXME: how to handle not scaled input?
    return values.to_array(dtype=dtype)


# Register as standard JAX primitive
stop_scaling_p.multiple_results = False
stop_scaling_p.def_abstract_eval(stop_scaling_abstract_eval)
stop_scaling_p.def_impl(stop_scaling_impl)
mlir.register_lowering(stop_scaling_p, stop_scaling_mlir_lowering)
# Register "scaled" translation.
register_scaled_op(stop_scaling_p, scaled_stop_scaling)


get_data_scale_p = core.Primitive("get_data_scale_p")
"""`get_data_scale` unbundling JAX primitive: return a tuple of data and scale
arrays.

In standard JAX, this is just an operation returning the input array and a constant scalar(1).

In JAX Scalify mode, it will return the pair of data and scale tensors
from a ScaledArray.
"""


def get_scale_dtype() -> Optional[DTypeLike]:
    """Get the scale dtype, if set in the Scalify config."""
    return get_scalify_config().scale_dtype


def get_data_scale(values: Array) -> Array:
    """`get_data_scale` primitive call method."""
    return get_data_scale_p.bind(values)


def get_data_scale_impl(values: Array) -> Array:
    if isinstance(values, ScaledArray):
        return (values.data, values.scale)
    # Use array dtype for scale by default.
    scale_dtype = get_scale_dtype() or values.dtype
    scale = np.ones((), dtype=scale_dtype)
    return values, scale


def get_data_scale_abstract_eval(values: core.ShapedArray) -> core.ShapedArray:
    if isinstance(values, ScaledArray):
        return (values.data, values.scale)
    # Use array dtype for scale by default.
    scale_dtype = get_scale_dtype() or values.dtype
    return values, core.ShapedArray((), dtype=scale_dtype)


def get_data_scale_mlir_lowering(
    ctx: LoweringRuleContext, *args: Union[ir.Value, Sequence[ir.Value]]
) -> Sequence[Union[ir.Value, Sequence[ir.Value]]]:
    # Just forwarding `values` term, adding a constant scalar scale(1).
    assert len(args) == 1
    assert len(ctx.avals_in) == 1
    assert len(ctx.avals_out) == 2
    # Scale dtype "decided" during initial JAX tracing.
    scale_dtype = ctx.avals_out[1].dtype
    scale = ir_constant(np.ones((), dtype=scale_dtype))
    return (args[0], scale)


def scaled_get_data_scale(values: ScaledArray) -> Array:
    """Scaled `get_data_scale` implementation: return scale tensor."""
    scale_dtype = get_scale_dtype()
    # Mis-match may potentially create issues (i.e. not equivalent scale dtype after scalify tracer)!
    if scale_dtype != values.scale.dtype:
        logging.warning(
            f"Scalify config scale dtype not matching ScaledArray scale dtype: '{values.scale.dtype}' vs '{scale_dtype}'. Scalify graph transformation may fail because of that."
        )
    return values.data, values.scale


# Register as standard JAX primitive
get_data_scale_p.multiple_results = True
get_data_scale_p.def_abstract_eval(get_data_scale_abstract_eval)
get_data_scale_p.def_impl(get_data_scale_impl)
mlir.register_lowering(get_data_scale_p, get_data_scale_mlir_lowering)
# Register "scaled" translation.
register_scaled_op(get_data_scale_p, scaled_get_data_scale)


def rebalance(values: ScaledArray, rebalance_scale: Array) -> ScaledArray:
    """Rebalance a ScaledArray with a scale component, i.e.
    respectively divide/multiply the data/scale by the rebalance scale.

    NOTE: no-op in normal JAX mode.

    Args:
        values: Array/ScaledArray to rebalance.
        rebalance_scale: Rebalancing scale value.
    Returns:
        Array/ScaledArray rebalanced.
    """
    _, scale = get_data_scale(values)
    assert scale.dtype == rebalance_scale.dtype
    out_scale = scale * rebalance_scale
    return set_scaling(values, out_scale)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from dataclasses import dataclass
from enum import IntEnum
from functools import partial, wraps
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union

import jax
import numpy as np
from jax import core, lax
from jax._src.custom_derivatives import (
    custom_jvp_call_jaxpr_p,
    custom_jvp_call_p,
    custom_vjp_call_jaxpr_p,
    custom_vjp_call_p,
)
from jax._src.util import safe_map
from jax.tree_util import register_pytree_node_class

from .datatype import (
    Array,
    ArrayTypes,
    DTypeLike,
    ScaledArray,
    Shape,
    as_scaled_array_base,
    is_scaled_leaf,
    is_static_zero,
)
from .pow2 import Pow2RoundMode
from .utils import python_scalar_as_numpy


@dataclass(frozen=True)
class ScalifyConfig:
    """Scalify configuration/parameters when tracing a graph.

    NOTE: this config can be locally changed using a Python context manager:
    `with ScalifyConfig(...):`

    Args:
        rounding_mode: Power-of-2 rounding mode.
        scale_dtype: Scale (default) datatype.
    """

    rounding_mode: Pow2RoundMode = Pow2RoundMode.DOWN
    scale_dtype: DTypeLike = None

    def __enter__(self):
        global _scalify_config_stack
        _scalify_config_stack.append(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _scalify_config_stack
        _scalify_config_stack.pop()


# Scalify config stack.
_scalify_config_stack = [ScalifyConfig()]


def get_scalify_config() -> ScalifyConfig:
    """Get current/local scalify config."""
    return _scalify_config_stack[-1]


class ScaledPrimitiveType(IntEnum):
    """Scale (JAX) primitive type.

    This enum described the behaviour when `scalify` is
    tracing the graph.

    FORWARD: Forwarding scaling => only used if scaled inputs.
        Default behaviour.
    ALWAYS_SCALE: Always use scaled version.
    """

    NEVER = 0
    FORWARD = 1
    ALWAYS_SCALE = 2


_scaled_jaxpr_ops_registry: Dict[core.Primitive, Any] = {}
"""Registry of (sub) "jaxpr" ops/primitives and their scaled translation.

The core "jaxpr" primitives are typical `pjit`, `xla_call`, where the JSA interpreter
will need to be run on sub-jaxprs, passing the full metadata on input/output tensors.
"""


_scaled_ops_registry: Dict[core.Primitive, Tuple[Any, ScaledPrimitiveType]] = {}
"""Registry of JAX common primitives and their scaled translation.
"""


_broadcasted_scalar_preserving_primitives: Set[core.Primitive] = set()
"""Scalar (broadcasted array) preserving JAX primitives

More specifically: if all inputs are (broadcasted) scalars, then the output(s)
are broadcasted scalars. Keeping track of broadcasted scalars is allowing
proper conversion to ScaledArrays (instead of assigning default scale 1).
"""

_broadcasted_zero_preserving_primitives: Set[core.Primitive] = set()
"""Zero (broadcasted array) preserving JAX primitives

More specifically: if all inputs are (broadcasted) zero scalars, then the output(s)
are broadcasted zero scalars. Keeping track of broadcasted zero scalars is allowing
proper conversion to ScaledArrays (instead of assigning default scale 1).
"""


def _get_lax_prim(scaled_func: Any) -> core.Primitive:
    try:
        prim_name = scaled_func.__name__.replace("scaled_", "") + "_p"
        prim = getattr(jax.lax, prim_name)
    except AttributeError:
        raise AttributeError(f"Could not find corresponding 'jax.lax' primitive for '{scaled_func.__name__}'.")
    # Check as well it is a proper primitive! And not something else also in `jax.lax`
    if not isinstance(prim, core.Primitive):
        raise AttributeError(f"The object `{prim}` is not a proper JAX primitive for '{scaled_func.__name__}'.")
    return prim


def _get_aval(val: Any) -> core.ShapedArray:
    """Get the abstract value (i.e. ShapedArray) from any input."""
    if hasattr(val, "aval"):
        return val.aval
    return core.ShapedArray(shape=val.shape, dtype=val.dtype)


def _get_data(val: Any) -> Array:
    if isinstance(val, ScaledArray):
        return val.data
    return val


def register_scaled_op(
    prim: core.Primitive, scaled_func: Any, scaled_type: ScaledPrimitiveType = ScaledPrimitiveType.FORWARD
) -> None:
    """Register the scaled translation of JAX primitive.

    Raises an error if a scaled translation is already existing for this primitive.

    Args:
        prim: JAX primitive.
        scaled_func: Scaled translation of the primitive. With the same interface.
        scaled_type: Scaled primitive type => behaviour when `scalify` tracing.
    """
    assert isinstance(prim, core.Primitive)
    # Can not register a jaxpr type op this way.
    assert prim not in _scaled_jaxpr_ops_registry
    if prim in _scaled_ops_registry:
        raise KeyError(f"A scaled translation is already registered for the JAX primitive '{prim}'.")
    _scaled_ops_registry[prim] = (scaled_func, scaled_type)


def register_scaled_lax_op(scaled_func):
    """
    Registers a scaled function/translation into the scaled_ops_registry by matching
    the function name with pattern `scaled_{func_name}` to a primitive in the
    `jax.lax` namespace.

    Example: `scaled_mul` is matched to `jax.lax.mul_p`
    """
    lax_prim = _get_lax_prim(scaled_func)
    register_scaled_op(lax_prim, scaled_func, ScaledPrimitiveType.FORWARD)
    # Always return the function in the case of decorator use.
    return scaled_func


def find_registered_scaled_op(prim: core.Primitive) -> Tuple[Any, ScaledPrimitiveType]:
    """Find a registered JAX scaled operation/translation. Returns (None, None) if
    the primitive does not have a scaled translation registered.

    Args:
        prim: JAX primitive.
    """
    return _scaled_ops_registry.get(prim, (None, ScaledPrimitiveType.NEVER))


@register_pytree_node_class
@dataclass(frozen=True, init=False)
class ScalifyTracerArray:
    """Meta-Array class used in scalify tracer. It can represent
    any array, scaled or not, and tracks whether an array corresponds to a scalar broadcasted.

    Compatible with JAX PyTrees in order to be able to trace a graph with `ScalifyTracerArray`
    as inputs/outputs.

    Args:
        array: Normal or scaled array.
        is_broadcasted_scalar: Is the array a broadcasted scalar (metadata).
        is_broadcasted_zero: Is the array a broadcased zero scalar (metadata).
    """

    array: Union[Array, ScaledArray] = None
    # Metadata fields.
    is_broadcasted_scalar: bool = False
    is_broadcasted_zero: bool = False

    def __init__(
        self,
        arr: Union[Array, ScaledArray],
        is_broadcasted_scalar: Optional[bool] = None,
        is_broadcasted_zero: Optional[bool] = None,
    ) -> None:
        # Convert Python scalars, if necessary.
        arr = python_scalar_as_numpy(arr)
        assert isinstance(arr, (np.bool_, np.number, np.ndarray, ScaledArray, *ArrayTypes))
        object.__setattr__(self, "array", arr)

        # Is a zero broadcasted scalar? Only checking when info is not provided.
        if is_broadcasted_zero is None:
            is_broadcasted_zero = bool(np.all(is_static_zero(self.array)))
        object.__setattr__(self, "is_broadcasted_zero", is_broadcasted_zero)
        # Optional is broadcasted scalar information (always consistent with broadcasted zero!)
        is_scalar = self.array.size == 1
        is_broadcasted_scalar = is_scalar if is_broadcasted_scalar is None else is_broadcasted_scalar or is_scalar
        is_broadcasted_scalar = is_broadcasted_scalar or is_broadcasted_zero
        object.__setattr__(self, "is_broadcasted_scalar", is_broadcasted_scalar)

        # Always make sure we have zero scale to represent broadcasted zero.
        if is_broadcasted_zero and isinstance(self.array, ScaledArray):
            object.__setattr__(self.array, "scale", np.array(0, self.array.scale.dtype))

    def tree_flatten(self):
        # See official JAX documentation on extending PyTrees.
        # Note: using explicit tree flatten instead of chex for MyPy compatibility.
        children = (self.array,)
        aux_data = (self.is_broadcasted_scalar, self.is_broadcasted_zero)
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        # See official JAX documentation on extending PyTrees.
        assert len(aux_data) == 2
        assert len(children) == 1
        return cls(children[0], is_broadcasted_scalar=aux_data[0], is_broadcasted_zero=aux_data[1])

    @property
    def dtype(self) -> DTypeLike:
        return self.array.dtype

    @property
    def size(self) -> int:
        return self.array.size

    @property
    def shape(self) -> Shape:
        return self.array.shape

    @property
    def is_scaled_array(self) -> bool:
        return isinstance(self.array, ScaledArray)

    def to_scaled_array(self, scale_dtype: Optional[DTypeLike] = None) -> ScaledArray:
        """(Tentatively) converting to a scaled array.

        Supporting the following cases:
            - scalar array;
            - broadcasted scalar array;

        Not supporting:
            - bool/int dtypes;
            - any other array;

        TODO: support (constant) Numpy arrays.
        """
        # Already scaled array, or not a floating point dtype.
        if isinstance(self.array, ScaledArray) or not np.issubdtype(self.dtype, np.floating):
            return self.array

        if self.is_broadcasted_zero:
            # Directly create the scaled array.
            scale_dtype = scale_dtype or self.array.dtype
            scale = np.array(0, dtype=scale_dtype)
            return ScaledArray(data=self.array, scale=scale)
        elif np.ndim(self.array) == 0:
            # Single value => "easy case".
            return as_scaled_array_base(self.array, scale_dtype=scale_dtype)
        elif self.is_broadcasted_scalar:
            # Broadcasted scalar => convert as a scalar.
            scalar_val = self.array.ravel()[0]
            scaled_scalar = as_scaled_array_base(scalar_val, scale_dtype=scale_dtype)
            return as_scaled_array_base(self.array, scale=scaled_scalar.scale)

        # No promotion rule found => just return as such.
        return self.array

    def to_array(self) -> Array:
        """Converting to a (normal) JAX/Numpy array."""
        if not self.is_scaled_array:
            return self.array
        return self.array.to_array()


def scalify(fun):
    """`scalify` JAX graph transformation.

    The `scalify` graph transformation works in a forwarding mode:
        scaled arrays are forwarded to scaled primitives, which will generate scaled outputs.

    If no inputs to a JAX primitive are scaled -> the normal primitive is then called, generating a common
    JAX output array.

    This behaviour is the standard one for `ScaledPrimitiveType.FORWARD` primitives.
    An alternative behaviour is possible for `ScaledPrimitiveType.ALWAYS_SCALED` primitives, where the scaled
    operation will always be called. A typical example is the `set_scaling` primitive.
    """

    @wraps(fun)
    def wrapped(*args, **kwargs):
        if len(kwargs) > 0:
            raise NotImplementedError("`scalify` JAX interpreter not supporting named tensors at present.")

        aval_args = jax.tree_util.tree_map(_get_aval, args, is_leaf=is_scaled_leaf)
        # Get jaxpr of unscaled/normal graph. Getting output Pytree shape as well.
        closed_jaxpr, outshape = jax.make_jaxpr(fun, return_shape=True)(*aval_args, **kwargs)
        out_leaves, out_pytree = jax.tree_util.tree_flatten(outshape)

        # Flattening of PyTree inputs.
        inputs_scaled = args
        inputs_scaled_flat, _ = jax.tree_util.tree_flatten(inputs_scaled, is_leaf=is_scaled_leaf)
        # Convert to Scalify tracer (meta) arrays.
        inputs_tracer_flat = list(map(ScalifyTracerArray, inputs_scaled_flat))
        consts_tracer_flat = list(map(ScalifyTracerArray, closed_jaxpr.literals))
        # Trace the graph & convert to scaled one.
        outputs_tracer_flat = scalify_jaxpr(closed_jaxpr.jaxpr, consts_tracer_flat, *inputs_tracer_flat)
        outputs_scaled_flat = [v.array for v in outputs_tracer_flat]
        # Reconstruct the output Pytree, with scaled arrays.
        # NOTE: this step is also handling single vs multi outputs.
        assert len(out_leaves) == len(outputs_scaled_flat)
        output_scaled = jax.tree_util.tree_unflatten(out_pytree, outputs_scaled_flat)
        return output_scaled

    return wrapped


def jaxpr_eqn_bind(eqn: core.JaxprEqn, invals: Sequence[core.ShapedArray]) -> Sequence[core.ShapedArray]:
    """Bind a Jaxpr equation to arrays."""
    subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
    outvals = eqn.primitive.bind(*subfuns, *invals, **bind_params)
    if not eqn.primitive.multiple_results:
        outvals = [outvals]
    return outvals


def scalify_jaxpr(
    jaxpr: core.Jaxpr, consts: Sequence[ScalifyTracerArray], *args: ScalifyTracerArray
) -> Sequence[ScalifyTracerArray]:
    env: Dict[core.Var, ScalifyTracerArray] = {}
    # Check dtype consistency between normal and scaled modes.
    safe_check_dtypes: bool = False
    # Scalify config to use.
    scalify_cfg = get_scalify_config()

    def read(var: core.Var) -> ScalifyTracerArray:
        if type(var) is core.Literal:
            # Wrap the constant in tracer array.
            return ScalifyTracerArray(var.val)
        return env[var]

    def write(var: core.Var, val: ScalifyTracerArray) -> None:
        env[var] = val

    # A few initial checks to make sure there is consistency.
    assert len(jaxpr.invars) == len(args)
    assert len(jaxpr.constvars) == len(consts)
    safe_map(write, jaxpr.invars, args)
    safe_map(write, jaxpr.constvars, consts)

    for eqn in jaxpr.eqns:
        invals_tracer: List[ScalifyTracerArray] = safe_map(read, eqn.invars)
        if eqn.primitive in _scaled_jaxpr_ops_registry:
            # Core sub-jaxpr primitive => pass the complete tracer array with metadata.
            scaled_jaxpr_prim_fn = _scaled_jaxpr_ops_registry[eqn.primitive]
            outvals_tracer = scaled_jaxpr_prim_fn(*invals_tracer, **eqn.params)
            # Save outputs and move on!
            safe_map(write, eqn.outvars, outvals_tracer)
            continue

        # Common (scaled) JAX primitives path.
        # Is there any ScaledArray among inputs?
        any_scaled_inputs = any([v.is_scaled_array for v in invals_tracer])
        # Is there a scaled primitive associated?
        scaled_prim_fn, scaled_prim_type = _scaled_ops_registry.get(eqn.primitive, (None, ScaledPrimitiveType.NEVER))

        # Are outputs broadcasted scalars?
        are_inputs_broadcasted_scalars = all([v.is_broadcasted_scalar for v in invals_tracer])
        are_outputs_broadcasted_scalars = (
            are_inputs_broadcasted_scalars and eqn.primitive in _broadcasted_scalar_preserving_primitives
        )
        # Are outputs broadcasted zeroes?
        are_inputs_broadcasted_zeroes = all([v.is_broadcasted_zero for v in invals_tracer])
        are_outputs_broadcasted_zeroes = (
            are_inputs_broadcasted_zeroes and eqn.primitive in _broadcasted_zero_preserving_primitives
        )
        # Outputs scalify factory method.
        scalify_array_init_fn = lambda v: ScalifyTracerArray(
            v, is_broadcasted_scalar=are_outputs_broadcasted_scalars, is_broadcasted_zero=are_outputs_broadcasted_zeroes
        )

        if not any_scaled_inputs and scaled_prim_type != ScaledPrimitiveType.ALWAYS_SCALE:
            # Using normal JAX primitive: no scaled inputs, and not always scale rule.
            invals = [v.to_array() for v in invals_tracer]
            outvals = jaxpr_eqn_bind(eqn, invals)
            outvals_tracer = list(map(scalify_array_init_fn, outvals))
        elif scaled_prim_fn is None:
            raise NotImplementedError(
                f"'{eqn.primitive}' JAX primitive does not have an implementation for ScaledArray inputs yet."
            )
        else:
            # Using scaled primitive. Automatic promotion of inputs to scaled array, when possible.
            scaled_invals = [v.to_scaled_array(scalify_cfg.scale_dtype) for v in invals_tracer]
            outvals = scaled_prim_fn(*scaled_invals, **eqn.params)
            if not eqn.primitive.multiple_results:
                outvals = [outvals]
            outvals_tracer = list(map(scalify_array_init_fn, outvals))

            # Check consistency with normal JAX mode. Help catching dtype promotion errors.
            # NOTE: ignoring when no outputs! (e.g. debug_callback).
            if safe_check_dtypes and len(outvals) > 0:
                ref_outvals = jaxpr_eqn_bind(eqn, [_get_data(v.array) for v in invals_tracer])
                data_outvals = [_get_data(v) for v in outvals]
                # Check scaled dtypes == ref dtypes.
                ref_dtypes = tuple(v.dtype for v in ref_outvals)
                data_dtypes = tuple(v.dtype for v in data_outvals)
                if data_dtypes != ref_dtypes:
                    raise ValueError(
                        f"Output dtype of '{eqn.primitive}' scaled translation is not consistent with the JAX reference primitive implementation: {data_dtypes} vs {ref_dtypes}."
                    )

        safe_map(write, eqn.outvars, outvals_tracer)

    outvals_tracer = safe_map(read, jaxpr.outvars)
    return outvals_tracer


def scaled_pjit_translation(*args: ScalifyTracerArray, **kwargs: Any) -> Sequence[ScalifyTracerArray]:
    """Scaled translation of `pjit`. Basically re-running `scalify` on sub-jaxpr.

    NOTE: the `pjit` call will be kept, forwarding the proper parameters (shardings, ...).
    """
    closed_jaxpr = kwargs["jaxpr"]
    name = kwargs["name"]
    inline = kwargs["inline"]
    keep_unused = kwargs["keep_unused"]
    # TODO: properly adapt + pass these options.
    # donated_invars = kwargs["donated_invars"]
    # in_shardings = kwargs["in_shardings"]
    # out_shardings = kwargs["out_shardings"]

    consts_tracer_flat = [ScalifyTracerArray(v) for v in closed_jaxpr.literals]
    # Generate the sub-scaled function, with proper `jax.jit` options.
    subfunc = partial(scalify_jaxpr, closed_jaxpr.jaxpr, consts_tracer_flat)
    subfunc.__name__ = name  # type:ignore
    subfunc = jax.jit(subfunc, inline=inline, keep_unused=keep_unused)
    outvals = subfunc(*args)
    return outvals


try:
    from jax._src.pjit import pjit_p

    _scaled_jaxpr_ops_registry[pjit_p] = scaled_pjit_translation
except (ImportError, ModuleNotFoundError):
    pass


def scaled_xla_call_translation(*args: ScalifyTracerArray, **kwargs: Any) -> Sequence[ScalifyTracerArray]:
    """Scaled translation of `xla_call`. Basically re-running `scalify` on sub-jaxpr.

    Useful for JAX 0.3 compatibility
    """
    jaxpr = kwargs["call_jaxpr"]
    name = kwargs["name"]
    inline = kwargs["inline"]
    keep_unused = kwargs["keep_unused"]
    # TODO: properly adapt + pass these options.
    # donated_invars = kwargs["donated_invars"]
    # in_shardings = kwargs["in_shardings"]
    # out_shardings = kwargs["out_shardings"]

    assert len(jaxpr.constvars) == 0
    # Generate the sub-scaled function, with proper `jax.jit` options.
    subfunc = partial(scalify_jaxpr, jaxpr, [])
    subfunc.__name__ = name  # type:ignore
    subfunc = jax.jit(subfunc, inline=inline, keep_unused=keep_unused)
    outputs_scaled_flat = subfunc(*args)
    return outputs_scaled_flat


try:
    from jax.interpreters.xla import xla_call_p

    _scaled_jaxpr_ops_registry[xla_call_p] = scaled_xla_call_translation
except (ImportError, ModuleNotFoundError):
    pass


def scaled_custom_jvp_call_translation(*args: ScalifyTracerArray, **params: Any) -> Sequence[ScalifyTracerArray]:
    """Scaled translation of `custom_jvp_call` primitive. Forwarding the scaled call to sub-jaxpr,
    and modifying the underlying `jvp` function.
    """
    # [fun, jvp], bind_params = custom_jvp_call_p.get_bind_params(params)
    key_jaxpr = "call_jaxpr" if jax.__version_info__[1] > 3 else "fun_jaxpr"
    call_closed_jaxpr = params[key_jaxpr]
    # JAX 0.3 compatibility.
    assert params.get("num_consts", 0) == 0
    # FIXME: re-call the custom_jvp decorator/bind.
    call_subfunc = partial(scalify_jaxpr, call_closed_jaxpr.jaxpr, call_closed_jaxpr.literals)
    return call_subfunc(*args)


_scaled_jaxpr_ops_registry[custom_jvp_call_p] = scaled_custom_jvp_call_translation
_scaled_jaxpr_ops_registry[custom_jvp_call_jaxpr_p] = scaled_custom_jvp_call_translation


def scaled_custom_vjp_call_translation(*args: ScalifyTracerArray, **params: Any) -> Sequence[ScalifyTracerArray]:
    """Scaled translation of `custom_vjp_call` primitive. Forwarding the scaled call to sub-jaxpr,
    and modifying the underlying `vjp` function.
    """
    key_jaxpr = "fun_jaxpr"
    call_closed_jaxpr = params[key_jaxpr]
    # FIXME: re-call the custom_vjp decorator/bind.
    call_subfunc = partial(scalify_jaxpr, call_closed_jaxpr.jaxpr, call_closed_jaxpr.literals)
    return call_subfunc(*args)


_scaled_jaxpr_ops_registry[custom_vjp_call_p] = scaled_custom_vjp_call_translation
_scaled_jaxpr_ops_registry[custom_vjp_call_jaxpr_p] = scaled_custom_vjp_call_translation


# Default collection of scalar preserving JAX primitives.
_broadcasted_scalar_preserving_primitives |= {
    lax.abs_p,
    lax.acos_p,
    lax.acosh_p,
    lax.add_p,
    lax.asin_p,
    lax.asinh_p,
    lax.atan_p,
    lax.atan2_p,
    lax.atanh_p,
    lax.bitcast_convert_type_p,
    lax.broadcast_in_dim_p,
    lax.cbrt_p,
    lax.ceil_p,
    lax.clamp_p,
    lax.convert_element_type_p,
    lax.exp_p,
    lax.expm1_p,
    lax.floor_p,
    lax.integer_pow_p,
    lax.min_p,
    lax.max_p,
    lax.mul_p,
    lax.neg_p,
    lax.reduce_prod_p,
    lax.reduce_sum_p,
    lax.reduce_max_p,
    lax.reduce_min_p,
    lax.reduce_precision_p,
    lax.reshape_p,
    lax.rem_p,
    lax.slice_p,
    lax.sin_p,
    lax.sinh_p,
    lax.sub_p,
    lax.sqrt_p,
    lax.tan_p,
    lax.tanh_p,
    lax.transpose_p,
}

# Default collection of zero arrays preserving JAX primitives.
_broadcasted_zero_preserving_primitives |= {
    lax.abs_p,
    lax.add_p,
    lax.broadcast_in_dim_p,
    lax.cbrt_p,
    lax.ceil_p,
    lax.clamp_p,
    lax.convert_element_type_p,
    lax.exp_p,
    lax.expm1_p,
    lax.floor_p,
    lax.min_p,
    lax.max_p,
    lax.mul_p,
    lax.neg_p,
    lax.reduce_prod_p,
    lax.reduce_sum_p,
    lax.reduce_max_p,
    lax.reduce_min_p,
    lax.reduce_precision_p,
    lax.reshape_p,
    lax.slice_p,
    lax.sin_p,
    lax.sinh_p,
    lax.sub_p,
    lax.sqrt_p,
    lax.tan_p,
    lax.tanh_p,
    lax.transpose_p,
}



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, Union

import jax
import jax.numpy as jnp
import numpy as np
from chex import Shape
from jax.core import ShapedArray
from jax.tree_util import register_pytree_node_class
from numpy.typing import ArrayLike, DTypeLike, NDArray

from .pow2 import Pow2RoundMode, pow2_decompose
from .typing import Array, ArrayTypes

if TYPE_CHECKING:
    GenericArray = Union[Array, np.ndarray[Any, Any]]
else:
    GenericArray = Union[Array, np.ndarray]


@register_pytree_node_class
@dataclass
class ScaledArray:
    """ScaledArray: dataclass associating data and a scale.

    JAX Scaled Arithmetics provides a consistent JAX LAX implementation
    propagating scaling for low-precision arithmetics.

    Semantics: `ScaledArray` represents an array with the following values:
        self.data * self.scale
    where `self.scale` is always assumed to be broadcastable to `self.data`.

    Notes:
        1. Current implementation only supports `scale` being a scalar.
        2. `data` and `scale` can have different dtypes. `data` dtype is used as the
            reference dtype. Meaning a power of 2 `scale` is just a dtype `E8M0` for instance.

    Args:
        data: Un-scaled data array.
        scale: Scale array (scalar only supported at the moment).
            If `scale` is None, equivalent to a normal array.
    """

    data: GenericArray
    scale: GenericArray

    def __post_init__(self):
        # Always have a Numpy array as `data`.
        if isinstance(self.data, np.number):
            object.__setattr__(self, "data", np.array(self.data))
        # TODO/FIXME: support number as data?
        assert isinstance(self.data, (*ArrayTypes, np.ndarray))
        assert isinstance(self.scale, (*ArrayTypes, np.ndarray, np.number))
        # Only supporting scale scalar for now.
        assert self.scale.shape == ()

    def tree_flatten(self):
        # See official JAX documentation on extending PyTrees.
        # Note: using explicit tree flatten instead of chex for MyPy compatibility.
        children = (self.data, self.scale)
        return (children, None)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        # See official JAX documentation on extending PyTrees.
        assert len(children) == 2
        return cls(children[0], children[1])

    @property
    def dtype(self) -> DTypeLike:
        return self.data.dtype

    @property
    def shape(self) -> Shape:
        return self.data.shape

    @property
    def size(self) -> int:
        return self.data.size

    def to_array(self, dtype: DTypeLike = None) -> GenericArray:
        """Convert to the scaled array to a Numpy/JAX array.

        Args:
            dtype: Optional conversion dtype. `data.dtype` by default.
        """
        dtype = self.data.dtype if dtype is None else dtype
        data = self.data.astype(dtype)
        scale = self.scale.astype(dtype)
        values = data * scale
        return values

    def __array__(self, dtype: DTypeLike = None) -> NDArray[Any]:
        """Numpy array interface support."""
        return np.asarray(self.to_array(dtype))

    @property
    def aval(self) -> ShapedArray:
        """Abstract value of the scaled array, i.e. shape and dtype."""
        return ShapedArray(self.data.shape, self.data.dtype)

    def astype(self, dtype: DTypeLike) -> "ScaledArray":
        """Convert the ScaledArray to a dtype.
        NOTE: only impacting `data` field, not the `scale` tensor.
        """
        return ScaledArray(self.data.astype(dtype), self.scale)


def make_scaled_scalar(val: Array, scale_dtype: Optional[DTypeLike] = None) -> ScaledArray:
    """Make a scaled scalar (array), from a single value.

    The returned scalar will always be built such that:
        - data is scalar in [1, 2)
        - scale is a power-of-2 value.

    NOTE: data is chosen in [1, 2) instead of [0, 1) in order to
    keep any value representable in the same dtype, without overflowing.

    NOTE bis: only supporting floating point input.
    """
    # FIXME: implicit conversion from float64 to float32???
    if isinstance(val, float):
        val = np.float32(val)
    assert np.ndim(val) == 0
    assert np.issubdtype(val.dtype, np.floating)
    # Scale dtype to use. TODO: check the scale dtype is valid?
    scale_dtype = scale_dtype or val.dtype
    # Split mantissa and exponent in data and scale components.
    scale, mantissa = pow2_decompose(val, scale_dtype=scale_dtype, mode=Pow2RoundMode.DOWN)
    return ScaledArray(mantissa, scale)


def is_scaled_leaf(val: Any) -> bool:
    """Is input a normal JAX PyTree leaf (i.e. `Array`) or `ScaledArray1.

    This function is useful for JAX PyTree handling with `jax.tree` methods where
    the user wants to keep the ScaledArray data structures (i.e. not flattened as a
    pair of arrays).

    See `jax_scalify.tree` for PyTree `jax.tree` methods compatible with `ScaledArray`.
    """
    # TODO: check Numpy scalars as well?
    return np.isscalar(val) or isinstance(val, (Array, np.ndarray, ScaledArray))


def scaled_array_base(data: ArrayLike, scale: ArrayLike, dtype: DTypeLike = None, npapi: Any = jnp) -> ScaledArray:
    """ScaledArray (helper) base factory method, similar to `(j)np.array`."""
    data = npapi.asarray(data, dtype=dtype)
    scale = npapi.asarray(scale)
    return ScaledArray(data, scale)


def scaled_array(data: ArrayLike, scale: ArrayLike, dtype: DTypeLike = None, npapi: Any = jnp) -> ScaledArray:
    """ScaledArray (helper) factory method, similar to `(j)np.array`.

    Args:
        data: Main data/values.
        scale: Scale tensor.
        dtype: Optional dtype to use for the data.
        npapi: Numpy API to use.
    Returns:
        Scaled array instance.
    """
    return scaled_array_base(data, scale, dtype, npapi)


def as_scaled_array_base(
    val: Any, scale: Optional[ArrayLike] = None, scale_dtype: Optional[DTypeLike] = None
) -> Union[Array, ScaledArray]:
    """ScaledArray (helper) base factory method, similar to `(j)np.array`.

    Args:
        val: Value to convert to scaled array.
        scale: Optional scale value.
        scale_dtype: Optional (default) scale dtype.
    """
    if isinstance(val, ScaledArray):
        return val

    assert scale is None or scale_dtype is None
    # Simple case => when can ignore the scaling factor (i.e. 1 implicitely).
    is_static_one_scale: bool = scale is None or is_static_one_scalar(scale)  # type:ignore
    # Trivial cases: bool, int, float.
    if is_static_one_scale and isinstance(val, (bool, int)):
        return val
    if is_static_one_scale and isinstance(val, float):
        return make_scaled_scalar(np.float32(val), scale_dtype)

    # Ignored dtypes by default: int and bool
    ignored_dtype = np.issubdtype(val.dtype, np.integer) or np.issubdtype(val.dtype, np.bool_)
    if ignored_dtype:
        return val
    # Floating point scalar
    if val.ndim == 0 and is_static_one_scale:
        return make_scaled_scalar(val, scale_dtype)

    scale_dtype = scale_dtype or val.dtype
    scale = np.array(1, dtype=scale_dtype) if scale is None else scale
    if isinstance(val, (np.ndarray, *ArrayTypes)):
        if is_static_one_scale:
            return ScaledArray(val, scale)
        else:
            return ScaledArray(val / scale.astype(val.dtype), scale)  # type:ignore

    # TODO: fix bug when scale is not 1.
    raise NotImplementedError(f"Constructing `ScaledArray` from {val} and {scale} not supported.")  # type:ignore
    # return scaled_array_base(val, scale)


def as_scaled_array(val: Any, scale: Optional[ArrayLike] = None) -> ScaledArray:
    """ScaledArray (helper) factory method, similar to `(j)np.array`.

    NOTE: by default, int and bool values/arrays will be returned unchanged, as
    in most cases, there is no value representing these as scaled arrays.

    Compatible with JAX PyTree.

    Args:
        val: Main data/values or existing ScaledArray.
        scale: Optional scale to use when (potentially) converting.
    Returns:
        Scaled array instance.
    """
    return jax.tree_util.tree_map(lambda x: as_scaled_array_base(x, scale), val, is_leaf=is_scaled_leaf)


def asarray_base(val: Any, dtype: DTypeLike = None) -> GenericArray:
    """Convert back to a common JAX/Numpy array, base function."""
    if isinstance(val, ScaledArray):
        return val.to_array(dtype=dtype)
    elif isinstance(val, (*ArrayTypes, np.ndarray)):
        if dtype is None:
            return val
        return val.astype(dtype=dtype)
    # Convert to Numpy all other cases?
    return np.asarray(val, dtype=dtype)


def asarray(val: Any, dtype: DTypeLike = None) -> GenericArray:
    """Convert back to a common JAX/Numpy array.

    Compatible with JAX PyTree.

    Args:
        dtype: Optional dtype of the final array.
    """
    return jax.tree_util.tree_map(lambda x: asarray_base(x, dtype), val, is_leaf=is_scaled_leaf)


def is_numpy_scalar_or_array(val):
    return isinstance(val, np.ndarray) or np.isscalar(val)


def is_static_zero(val: Union[Array, ScaledArray]) -> Array:
    """Is a scaled array a static zero value (i.e. zero during JAX tracing as well)?

    Returns a boolean Numpy array of the shape of the input.
    """
    if is_numpy_scalar_or_array(val):
        return np.equal(val, 0)
    if isinstance(val, ScaledArray):
        data_mask = (
            np.equal(val.data, 0) if is_numpy_scalar_or_array(val.data) else np.zeros(val.data.shape, dtype=np.bool_)
        )
        scale_mask = (
            np.equal(val.scale, 0) if is_numpy_scalar_or_array(val.scale) else np.zeros(val.scale.shape, dtype=np.bool_)
        )
        return np.logical_or(data_mask, scale_mask)
    # By default: can't decide.
    return np.zeros(val.shape, dtype=np.bool_)


def is_static_anyscale(val: Union[Array, ScaledArray]) -> Array:
    """Is a scaled array a static anyscale values (i.e. 0/inf/-inf during JAX tracing as well)?

    Returns a boolean Numpy array of the shape of the input.
    """

    def np_anyscale(arr):
        # Check if 0, np.inf or -np.inf
        absarr = np.abs(arr)
        return np.logical_or(np.equal(absarr, 0), np.equal(absarr, np.inf))

    if is_numpy_scalar_or_array(val):
        return np_anyscale(val)
    if isinstance(val, ScaledArray):
        # TODO: deal with 0 * inf issue?
        data_mask = (
            np_anyscale(val.data) if is_numpy_scalar_or_array(val.data) else np.zeros(val.data.shape, dtype=np.bool_)
        )
        scale_mask = (
            np_anyscale(val.scale) if is_numpy_scalar_or_array(val.scale) else np.zeros(val.scale.shape, dtype=np.bool_)
        )
        return np.logical_or(data_mask, scale_mask)
    # By default: can't decide.
    return np.zeros(val.shape, dtype=np.bool_)


def is_static_one_scalar(val: Array) -> Union[bool, np.bool_]:
    """Is a scaled array a static one scalar value (i.e. one during JAX tracing as well)?"""
    if isinstance(val, (int, float)):
        return val == 1
    elif is_numpy_scalar_or_array(val) and val.size == 1:
        return np.all(np.equal(val, 1))
    elif isinstance(val, ScaledArray) and val.size == 1:
        return is_static_one_scalar(val.data) and is_static_one_scalar(val.scale)
    return False


def get_scale_dtype(val: Any) -> DTypeLike:
    """Get the scale dtype. Compatible with arrays and scaled arrays."""
    if isinstance(val, ScaledArray):
        return val.scale.dtype
    return val.dtype



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import numpy as np

from jax_scalify.core import scaled_array, scalify
from jax_scalify.ops import debug_print


def test__debug_print__scaled_arrays(capfd):
    fmt = "INPUTS: {} + {}"

    def debug_print_fn(x):
        debug_print(fmt, x, x)

    input_scaled = scaled_array([2, 3], 2.0, dtype=np.float32)
    scalify(debug_print_fn)(input_scaled)
    # Check captured stdout and stderr!
    captured = capfd.readouterr()
    assert len(captured.err) == 0
    assert captured.out.strip() == fmt.format(input_scaled, input_scaled)



# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MNIST example.

Library file which executes the training and evaluation loop for MNIST.
The data is loaded using tensorflow_datasets.
"""

# See issue #620.
# pytype: disable=wrong-keyword-args

import jax
import jax.numpy as jnp
import ml_collections
import numpy as np
import optax
import tensorflow_datasets as tfds
from absl import logging
from flax import linen as nn  # type:ignore

# from flax.metrics import tensorboard
from flax.training import train_state

import jax_scalify as jsa


class CNN(nn.Module):
    """A simple CNN model."""

    @nn.compact
    def __call__(self, x):
        x = nn.Conv(features=32, kernel_size=(3, 3))(x)
        x = nn.relu(x)
        x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = nn.Conv(features=64, kernel_size=(3, 3))(x)
        x = nn.relu(x)
        x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = x.reshape((x.shape[0], -1))  # flatten
        x = nn.Dense(features=256)(x)
        x = nn.relu(x)
        x = nn.Dense(features=10)(x)
        return x


@jax.jit
def apply_model(state, images, labels):
    """Computes gradients, loss and accuracy for a single batch."""

    def loss_fn(params):
        logits = state.apply_fn({"params": params}, images)
        one_hot = jax.nn.one_hot(labels, 10)
        loss = jnp.mean(optax.softmax_cross_entropy(logits=logits, labels=one_hot))
        return loss, logits

    grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
    (loss, logits), grads = grad_fn(state.params)
    accuracy = jnp.mean(jnp.argmax(logits, -1) == labels)
    return grads, loss, accuracy


@jax.jit
def update_model(state, grads):
    return state.apply_gradients(grads=grads)


@jax.jit
@jsa.scalify
def apply_and_update_model(state, batch_images, batch_labels):
    # Jitting together forward + backward + update.
    grads, loss, accuracy = apply_model(state, batch_images, batch_labels)
    state = update_model(state, grads)
    return state, loss, accuracy


def train_epoch(state, train_ds, batch_size, rng):
    """Train for a single epoch."""
    train_ds_size = len(train_ds["image"])
    steps_per_epoch = train_ds_size // batch_size

    perms = jax.random.permutation(rng, len(train_ds["image"]))
    perms = perms[: steps_per_epoch * batch_size]  # skip incomplete batch
    perms = perms.reshape((steps_per_epoch, batch_size))

    epoch_loss = []
    epoch_accuracy = []

    for perm in perms:
        batch_images = train_ds["image"][perm, ...]
        batch_labels = train_ds["label"][perm, ...]
        # Transform batch to ScaledArray
        batch_images = jsa.as_scaled_array(batch_images)
        # Apply & update stages in scaled mode.
        state, loss, accuracy = apply_and_update_model(state, batch_images, batch_labels)

        epoch_loss.append(np.asarray(loss))
        epoch_accuracy.append(np.asarray(accuracy))

    train_loss = np.mean(epoch_loss)
    train_accuracy = np.mean(epoch_accuracy)
    return state, train_loss, train_accuracy


def get_datasets():
    """Load MNIST train and test datasets into memory."""
    ds_builder = tfds.builder("mnist")
    ds_builder.download_and_prepare()
    train_ds = tfds.as_numpy(ds_builder.as_dataset(split="train", batch_size=-1))
    test_ds = tfds.as_numpy(ds_builder.as_dataset(split="test", batch_size=-1))
    train_ds["image"] = jnp.float32(train_ds["image"]) / 255.0
    test_ds["image"] = jnp.float32(test_ds["image"]) / 255.0
    return train_ds, test_ds


def create_train_state(rng, config):
    """Creates initial `TrainState`."""
    cnn = CNN()
    params = cnn.init(rng, jnp.ones([1, 28, 28, 1]))["params"]
    tx = optax.sgd(config.learning_rate, config.momentum)
    return train_state.TrainState.create(apply_fn=cnn.apply, params=params, tx=tx)


def train_and_evaluate(config: ml_collections.ConfigDict, workdir: str) -> train_state.TrainState:
    """Execute model training and evaluation loop.

    Args:
      config: Hyperparameter configuration for training and evaluation.
      workdir: Directory where the tensorboard summaries are written to.

    Returns:
      The train state (which includes the `.params`).
    """
    train_ds, test_ds = get_datasets()
    rng = jax.random.key(0)

    # summary_writer = tensorboard.SummaryWriter(workdir)
    # summary_writer.hparams(dict(config))

    rng, init_rng = jax.random.split(rng)
    init_rng = jax.random.PRNGKey(1)

    state = create_train_state(init_rng, config)
    # Convert model & optimizer states to `ScaledArray``
    state = jsa.as_scaled_array(state)

    logging.info("Start Flax MNIST training...")

    for epoch in range(1, config.num_epochs + 1):
        rng, input_rng = jax.random.split(rng)
        state, train_loss, train_accuracy = train_epoch(state, train_ds, config.batch_size, input_rng)
        # NOTE: running evaluation on the plain normal arrays.
        _, test_loss, test_accuracy = apply_model(jsa.asarray(state), test_ds["image"], test_ds["label"])

        logging.info(
            "epoch:% 3d, train_loss: %.4f, train_accuracy: %.2f, test_loss: %.4f,"
            " test_accuracy: %.2f"
            % (
                epoch,
                train_loss,
                train_accuracy * 100,
                test_loss,
                test_accuracy * 100,
            )
        )

        # summary_writer.scalar("train_loss", train_loss, epoch)
        # summary_writer.scalar("train_accuracy", train_accuracy, epoch)
        # summary_writer.scalar("test_loss", test_loss, epoch)
        # summary_writer.scalar("test_accuracy", test_accuracy, epoch)

    # summary_writer.flush()
    return state



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from . import core, lax, ops, tree
from ._version import __version__
from .core import (  # noqa: F401
    Pow2RoundMode,
    ScaledArray,
    ScalifyConfig,
    as_scaled_array,
    asarray,
    debug_callback,
    scaled_array,
    scalify,
)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from typing import Any, Tuple

# import chex
import jax
import jax.numpy as jnp
import numpy as np

# Type aliasing. To be compatible with JAX 0.3 as well.
try:
    from jax import Array

    ArrayTypes: Tuple[Any, ...] = (Array,)
except ImportError:
    from jaxlib.xla_extension import DeviceArray as Array

    # Older version of JAX <0.4
    ArrayTypes = (Array, jax.interpreters.partial_eval.DynamicJaxprTracer)

try:
    from jax.stages import ArgInfo

    # Additional ArgInfo in recent JAX versions.
    ArrayTypes = (*ArrayTypes, ArgInfo)
except ImportError:
    pass


def get_numpy_api(val: Any) -> Any:
    """Get the Numpy API corresponding to an array.

    Using the NumPy API whenever possible when tracing a JAX graph
    allows for simple constant folding optimization.

    JAX or classic Numpy supported.
    """
    if isinstance(val, (np.ndarray, np.number)):
        return np
    if isinstance(val, ArrayTypes):
        return jnp
    raise NotImplementedError(f"Unsupported input type '{type(val)}'. No matching Numpy API.")



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from .cast import (  # noqa: F401
    cast_on_backward,
    cast_on_forward,
    reduce_precision_on_backward,
    reduce_precision_on_forward,
)
from .debug import debug_callback, debug_callback_grad, debug_print, debug_print_grad  # noqa: F401
from .rescaling import (  # noqa: F401
    dynamic_rescale_l1,
    dynamic_rescale_l1_grad,
    dynamic_rescale_l2,
    dynamic_rescale_l2_grad,
    dynamic_rescale_max,
    dynamic_rescale_max_grad,
)



# Copyright (c) 2024 Graphcore Ltd. All rights reserved.
"""A basic MNIST MLP training example using Flax and Optax.

Similar to JAX MNIST from scratch, but using Flax and Optax libraries.

This example aim is to show how Scalify can integrate with common
NN libraries such as Flax and Optax.
"""
import time
from functools import partial

import datasets
import jax
import jax.numpy as jnp
import numpy as np
import optax
from flax import linen as nn  # type:ignore

import jax_scalify as jsa

# from jax.scipy.special import logsumexp


def logsumexp(a, axis=None, keepdims=False):
    from jax import lax

    dims = (axis,)
    amax = jnp.max(a, axis=dims, keepdims=keepdims)
    # FIXME: not proper scale propagation, introducing NaNs
    # amax = lax.stop_gradient(lax.select(jnp.isfinite(amax), amax, lax.full_like(amax, 0)))
    amax = lax.stop_gradient(amax)
    out = lax.sub(a, amax)
    out = lax.exp(out)
    out = lax.add(lax.log(jnp.sum(out, axis=dims, keepdims=keepdims)), amax)
    return out


class MLP(nn.Module):
    """A simple 3 layers MLP model."""

    @nn.compact
    def __call__(self, x):
        x = nn.Dense(features=512, use_bias=True)(x)
        x = nn.relu(x)
        x = nn.Dense(features=512, use_bias=True)(x)
        x = nn.relu(x)
        x = nn.Dense(features=10, use_bias=True)(x)
        logprobs = x - logsumexp(x, axis=1, keepdims=True)
        return logprobs


def loss(model, params, batch):
    inputs, targets = batch
    preds = model.apply(params, inputs)
    # targets = jsa.lax.rebalance(targets, np.float32(1 / 8))
    return -jnp.mean(jnp.sum(preds * targets, axis=1))


def accuracy(model, params, batch):
    inputs, targets = batch
    target_class = jnp.argmax(targets, axis=1)
    preds = model.apply(params, inputs)
    predicted_class = jnp.argmax(preds, axis=1)
    return jnp.mean(predicted_class == target_class)


def update(model, optimizer, model_state, opt_state, batch):
    grads = jax.grad(partial(loss, model))(model_state, batch)
    # Optimizer update (state & gradients).
    updates, opt_state = optimizer.update(grads, opt_state, model_state)
    model_state = optax.apply_updates(model_state, updates)
    return model_state, opt_state


if __name__ == "__main__":
    step_size = 0.001
    num_epochs = 10
    batch_size = 128
    key = jax.random.PRNGKey(42)
    use_scalify: bool = True

    training_dtype = np.dtype(np.float16)
    optimizer_dtype = np.dtype(np.float16)
    scale_dtype = np.float32

    train_images, train_labels, test_images, test_labels = datasets.mnist()
    num_train = train_images.shape[0]
    num_complete_batches, leftover = divmod(num_train, batch_size)
    num_batches = num_complete_batches + bool(leftover)
    mnist_img_size = train_images.shape[-1]

    def data_stream():
        rng = np.random.RandomState(0)
        while True:
            perm = rng.permutation(num_train)
            for i in range(num_batches):
                batch_idx = perm[i * batch_size : (i + 1) * batch_size]
                yield train_images[batch_idx], train_labels[batch_idx]

    # Build model & initialize model parameters.
    model = MLP()
    model_state = model.init(key, np.zeros((batch_size, mnist_img_size), dtype=training_dtype))
    # Optimizer & optimizer state.
    # opt = optax.sgd(learning_rate=step_size)
    opt = optax.adam(learning_rate=step_size, eps=2**-16)
    opt_state = opt.init(model_state)
    # Freeze model, optimizer (with step size).
    update_fn = partial(update, model, opt)

    if use_scalify:
        # Transform parameters to `ScaledArray`.
        model_state = jsa.as_scaled_array(model_state, scale=scale_dtype(1.0))
        opt_state = jsa.as_scaled_array(opt_state, scale=scale_dtype(0.0001))
        # Scalify the update function as well.
        update_fn = jsa.scalify(update_fn)
    # Convert the model state (weights) & optimizer state to proper dtype.
    model_state = jsa.tree.astype(model_state, training_dtype)
    opt_state = jsa.tree.astype(opt_state, optimizer_dtype, floating_only=True)

    print(f"Using Scalify: {use_scalify}")
    print(f"Training data format: {training_dtype.name}")
    print(f"Optimizer data format: {optimizer_dtype.name}")
    print("")

    update_fn = jax.jit(update_fn)

    batches = data_stream()
    for epoch in range(num_epochs):
        start_time = time.time()

        for _ in range(num_batches):
            batch = next(batches)
            # Scaled micro-batch + training dtype cast.
            batch = jsa.tree.astype(batch, training_dtype)
            if use_scalify:
                batch = jsa.as_scaled_array(batch, scale=scale_dtype(1))
            with jsa.ScalifyConfig(rounding_mode=jsa.Pow2RoundMode.DOWN, scale_dtype=scale_dtype):
                model_state, opt_state = update_fn(model_state, opt_state, batch)

        epoch_time = time.time() - start_time

        # Evaluation in normal/unscaled float32, for consistency.
        unscaled_params = jsa.asarray(model_state, dtype=np.float32)
        train_acc = accuracy(model, unscaled_params, (train_images, train_labels))
        test_acc = accuracy(model, unscaled_params, (test_images, test_labels))
        print(f"Epoch {epoch} in {epoch_time:0.2f} sec")
        print(f"Training set accuracy {train_acc:0.5f}")
        print(f"Test set accuracy {test_acc:0.5f}")



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import jax.numpy as jnp
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized
from numpy.typing import NDArray

from jax_scalify.core import Array, ScaledArray, ScalifyConfig, scaled_array, scalify
from jax_scalify.lax.base_scaling_primitives import (
    get_data_scale,
    rebalance,
    scaled_set_scaling,
    set_scaling,
    stop_scaling,
)


class SetScalingPrimitiveTests(chex.TestCase):
    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__set_scaling_primitive__scaled_array__eager_mode(self, npapi):
        values = scaled_array([-1.0, 2.0], 2.0, dtype=np.float16, npapi=npapi)
        scale = npapi.float16(4)
        output = set_scaling(values, scale)
        assert isinstance(output, ScaledArray)
        # No implicit promotion to JAX array when not necessary.
        assert isinstance(output.data, npapi.ndarray)
        assert isinstance(output.scale, (npapi.number, npapi.ndarray))
        npt.assert_equal(output.scale, npapi.float16(4))
        npt.assert_array_equal(output, values)

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.parameters(
        {"arr": np.array([-1.0, 2.0], dtype=np.float32)},
        {"arr": scaled_array([-1.0, 2.0], 1.0, dtype=np.float16)},
        {"arr": scaled_array([-1.0, 2.0], 0.0, dtype=np.float32)},
    )
    def test__set_scaling_primitive__zero_scaling(self, arr):
        def fn(arr, scale):
            return set_scaling(arr, scale)

        scale = np.array(0, dtype=arr.dtype)
        out = self.variant(scalify(fn))(arr, scale)
        assert isinstance(out, ScaledArray)
        npt.assert_array_almost_equal(out.scale, 0)
        npt.assert_array_almost_equal(out.data, 0)

    @chex.variants(with_jit=True, without_jit=True)
    def test__set_scaling_primitive__proper_result_without_scalify(self):
        def fn(arr, scale):
            return set_scaling(arr, scale)

        fn = self.variant(fn)
        arr = jnp.array([2, 3], dtype=np.float32)
        scale = jnp.array(4, dtype=np.float32)
        out = fn(arr, scale)
        # Scale input ignored => forward same array.
        assert isinstance(out, jnp.ndarray)
        npt.assert_array_equal(out, arr)

    @chex.variants(with_jit=True, without_jit=False)
    @parameterized.parameters(
        # Testing different combination of scaled/unscaled inputs.
        {"arr": np.array([-1.0, 2.0], dtype=np.float32), "scale": np.array(4.0, dtype=np.float32)},
        {"arr": np.array([-1.0, 2.0], dtype=np.float16), "scale": np.array(4.0, dtype=np.float32)},
        {"arr": scaled_array([-1.0, 2.0], 1.0, dtype=np.float16), "scale": np.array(4.0, dtype=np.float32)},
        {"arr": scaled_array([-1.0, 2.0], 2.0, dtype=np.float32), "scale": scaled_array(1.0, 4.0, dtype=np.float32)},
        {"arr": scaled_array([-1.0, 2.0], 2.0, dtype=np.float16), "scale": scaled_array(1.0, 4.0, dtype=np.float32)},
    )
    def test__set_scaling_primitive__proper_result_with_scalify(self, arr, scale):
        def fn(arr, scale):
            return set_scaling(arr, scale)

        fn = self.variant(scalify(fn))
        out = fn(arr, scale)
        # Unchanged output tensor, with proper dtype.
        assert isinstance(out, ScaledArray)
        assert out.dtype == arr.dtype
        npt.assert_array_equal(out.scale, scale)
        npt.assert_array_equal(out, arr)

    @parameterized.parameters(
        {"scale": 1},
        {"scale": np.int32(1)},
        {"scale": np.float32(1)},
    )
    def test__scaled_set_scaling__unchanged_scaled_array(self, scale):
        val = scaled_array([-1.0, 2.0], 2.0, dtype=np.float16)
        assert scaled_set_scaling(val, scale) is val

    @parameterized.parameters(
        {"scale": np.int32(1)},
        {"scale": np.float32(1)},
    )
    def test__scaled_set_scaling__unchanged_data_scaled_array(self, scale):
        val: NDArray[np.float16] = np.array([-1.0, 2.0], dtype=np.float16)
        out = scaled_set_scaling(val, scale)  # type:ignore
        assert isinstance(out, ScaledArray)
        assert out.data is val


class StopScalingPrimitiveTests(chex.TestCase):
    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__stop_scaling_primitive__scaled_array__eager_mode(self, npapi):
        values = scaled_array([-1.0, 2.0], 2.0, dtype=np.float16, npapi=npapi)
        output = stop_scaling(values)
        assert isinstance(output, npapi.ndarray)
        npt.assert_array_equal(output, values)

    @chex.variants(with_jit=True, without_jit=True)
    def test__stop_scaling_primitive__proper_result_without_scalify(self):
        def fn(arr):
            # Testing both variants.
            return stop_scaling(arr), stop_scaling(arr, dtype=np.float16)

        arr = jnp.array([2, 3], dtype=np.float32)
        out0, out1 = self.variant(fn)(arr)
        assert out0.dtype == arr.dtype
        assert out1.dtype == np.float16
        npt.assert_array_equal(out0, arr)
        npt.assert_array_almost_equal(out1, arr)

    @chex.variants(with_jit=True, without_jit=True)
    def test__stop_scaling_primitive__proper_result_with_scalify(self):
        def fn(arr):
            # Testing both variants.
            return stop_scaling(arr), stop_scaling(arr, dtype=np.float16)

        fn = self.variant(scalify(fn))
        arr = scaled_array([-1.0, 2.0], 3.0, dtype=np.float32)
        out0, out1 = fn(arr)
        assert isinstance(out0, Array)
        assert isinstance(out1, Array)
        assert out0.dtype == arr.dtype
        assert out1.dtype == np.float16
        npt.assert_array_equal(out0, arr)
        npt.assert_array_almost_equal(out1, arr)


class GetDataScalePrimitiveTests(chex.TestCase):
    @chex.variants(with_jit=True, without_jit=True)
    def test__get_data_scale_primitive__proper_result_without_scalify(self):
        def fn(arr):
            # Set a default scale dtype.
            with ScalifyConfig(scale_dtype=np.float32):
                return get_data_scale(arr)

        fn = self.variant(fn)
        arr = jnp.array([2, 3], dtype=np.float16)
        data, scale = fn(arr)
        assert data.dtype == np.float16
        assert scale.dtype == np.float32
        npt.assert_array_equal(data, arr)
        npt.assert_equal(scale, np.array(1, np.float32))

    @chex.variants(with_jit=True, without_jit=True)
    def test__get_data_scale_primitive__proper_result_with_scalify(self):
        def fn(arr):
            return get_data_scale(arr)

        fn = self.variant(scalify(fn))
        arr = scaled_array([2, 3], np.float16(4), dtype=np.float16)
        data, scale = fn(arr)
        npt.assert_array_equal(data, arr.data)
        npt.assert_equal(scale, arr.scale)

    def test__get_data_scale_primitive__numpy_input(self):
        arr = scaled_array([2, 3], 4, dtype=np.float16)
        # ScaledArray input.
        data, scale = get_data_scale(arr)
        npt.assert_array_equal(data, arr.data)
        npt.assert_array_equal(scale, arr.scale)
        # Normal numpy array input.
        data, scale = get_data_scale(np.asarray(arr))
        npt.assert_array_equal(data, arr)
        npt.assert_almost_equal(scale, 1)


class RebalancingOpsTests(chex.TestCase):
    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__rebalance_op__normal_array__eagermode(self, npapi):
        values = npapi.array([-1, 2], dtype=np.float16)
        output = rebalance(values, np.float16(2))
        # Same Python array object forwarded.
        assert output is values

    def test__rebalance_op__scaled_array__eagermode(self):
        arr_in = scaled_array([2, 3], np.float16(4), dtype=np.float16)
        scale_in = np.float16(0.5)
        arr_out = rebalance(arr_in, scale_in)
        assert isinstance(arr_out, ScaledArray)
        npt.assert_array_equal(arr_out.scale, arr_in.scale * scale_in)
        npt.assert_array_equal(arr_out, arr_in)
